/*
 perspective_test.c
  a thing

 Created: 2008.04.03 21:53 EST
 Revised: 2008.05.20 23:24 EST
*/

#include <cseries.h>
#include <cs_math.h>
#include <cs_graphics.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <MacWindows.h>
#include <QDOffscreen.h>
#include <Events.h>

/* ---- carbon solution macros */

#if CALL_NOT_IN_CARBON
# define InvalWindowRect(w,r)		do{GrafPtr _tmpp; GetPort(&_tmpp); \
	SetPortWindowPort(w); InvalRect(r); SetPort(_tmpp);}while(0)
#endif

/* ---- constants */

/*
#define MINIMUM_WORLD_WIDTH			128
#define MINIMUM_WORLD_HEIGHT		80

#define MAXIMUM_WORLD_WIDTH			384
#define MAXIMUM_WORLD_HEIGHT		192
*/

#define WORLD_WIDTH					384
#define WORLD_HEIGHT				192

#define BIT_DEPTH					16

#define DEFAULT_NORMAL				{255*1.0,255*0.5,255*0.5,255*0.5}

#define DEFAULT_TERMINAL_VELOCITY	(8192/6)

#define MINIMUM_ZOOM				256
#define MAXIMUM_ZOOM				65536
#define ZOOM_MAGNITUDE				64

enum
{ /* part identifiers */
	_no_part,
	_endpoint_part,
	_triangle_part,
	_player_part,
	_walking_constants_part,
	_running_constants_part
};

/* ---- structures / typedefs */

typedef struct world_point/*(12)*/
{
	SInt32			x, y, z;
} world_point;

typedef struct world_endpoint/*(16)*/
{
	SInt32			x, y, z, v; /* coordinate, identifier value */
} world_endpoint;

typedef struct basic_normal/*(4)*/
{
	UInt8			ca, cr, cg, cb; /* color */
} basic_normal;

typedef struct world_triangle/*(16+)*/
{
	SInt32			e0, e1, e2; /* endpoints */
	basic_normal	basic;
} world_triangle;

typedef struct acceleration_constants/*(32)*/
{
	SInt32			mxy, myz, mzx; /* turning magnitudes */

	SInt32			forward_magnitude; /* backward is little slower, sidestepping much slower */
	/* positional = forward/backward/sidestep */
	SInt32			positional_deceleration, turning_deceleration;
	SInt32			terminal_turning_magnitude, terminal_acceleration;
} acceleration_constants;

typedef struct world_player/*(160)*/
{
	SInt32			px, py, pz; /* position */
	SInt32			vx, vy, vz; /* velocity - acceleration adds to velocity */
	SInt32			ax, ay, az; /* acceleration - holding down keys accelerates */
	UInt32			xy, yz, zx; /* angular rotations */
/*0x30*/
	SInt32			vxy, vyz, vzx; /* turning accelerations */
	SInt32			sx, sy, sz, sd; /* scales; sd = 1.0 */

	SInt32			angular_acceleration; /* clipped to terminal acceleration */
/*0x50*/
	SInt32			angular_velocity; /* clipped to terminal velocity */

	UInt32			new_action_flags, old_action_flags;

	/* physics constants */
	SInt32			terminal_velocity;
/*0x60*/
	acceleration_constants	walking_constants, running_constants;
} world_player;

typedef struct world_struct/*(24+)*/
{
	SInt32			endpoint_count, triangle_count;
	UInt32			time_frame;
	UInt16			random_seed;
	UInt8			player_count, local_player; /* only one player for now */
/*0x10*/
	world_endpoint	*endpoints;
	world_triangle	*triangles;
	world_player	*players;
} world_struct;

typedef struct key_setting/*(4)*/
{
	UInt8			map_index, flags, char_code, char_index;
} key_setting, key_settings[32];

/* ---- globals */

char				temporary[1024];

key_settings		local_keys;

/* ---- prototypes */

void set_keys(key_settings*);

UInt32 ihypot2(SInt32, SInt32);
UInt32 ihypot3(SInt32, SInt32, SInt32);

/* ---- code */

#define MAXIMUM_LINE_READ_SIZE		1024
#define DBLSPC_CHAR					'\312'

#define vdispose(a,c)				do{c=0; if(a!=NULL) {free(a); a=NULL;}}while(0)

#define vpushback(a,c,x)			do{void *_tmpp=a; if(c<0) c=0; \
	if(((c&7)==0||a==NULL)&&(_tmpp=realloc(a,(c+8)*sizeof(*a)))==NULL) \
		{fputs("### vpushback: out of memory",stderr); goto cancel;} \
	a=_tmpp; a[c]=x; if(++c<=0){fputs("### vpushback: count overflow",stderr); \
	goto cancel;}}while(0)

static void dispose_world(world_struct *world)
{
	if (world==NULL)
		return;
	vdispose(world->endpoints,world->endpoint_count);
	vdispose(world->triangles,world->triangle_count);
	vdispose(world->players,world->player_count);
	free(world);
}

static SInt32 find_endpoint(world_endpoint *e, SInt32 c, SInt32 v)
{
	SInt32			x;

	if ((x=c)<=0)
		return -1;
	do
	{
		if (e->v==v)
			return c-x;
	} while(++e,--x);

	return -1;
}

/* returns true on success */
static Boolean reconnect_triangles(world_struct *world)
{
	world_endpoint	*endpoints = NULL;
	world_triangle	*triangle = NULL;
	SInt32			i;

	if (world==NULL || (endpoints=world->endpoints)==NULL
			|| (triangle=world->triangles)==NULL)
		{DebugStr("\preconnect_triangles: got NULL pointer"); return false;}
	for(i=0; i<world->triangle_count; ++i,++triangle)
	{
		if ((triangle->e0=find_endpoint(endpoints,world->endpoint_count,triangle->e0))<0)
			{fprintf(stderr,"### triangle #%u - e%u cannot be found\n", i,0); return false;}
		if ((triangle->e1=find_endpoint(endpoints,world->endpoint_count,triangle->e1))<0)
			{fprintf(stderr,"### triangle #%u - e%u cannot be found\n", i,1); return false;}
		if ((triangle->e2=find_endpoint(endpoints,world->endpoint_count,triangle->e2))<0)
			{fprintf(stderr,"### triangle #%u - e%u cannot be found\n", i,2); return false;}
	}

	return true;
}

#define COMMENT_CHAR				'#'

static world_struct *parse_world_file(FILE *input)
{
	char			line[MAXIMUM_LINE_READ_SIZE];
	basic_normal	default_normal = DEFAULT_NORMAL;

	world_struct	*world = NULL;
	world_endpoint	*endpoints = NULL;
	world_triangle	*triangles = NULL;
	world_player	*players = NULL;

	SInt32			line_count = 0;

	if ((world=malloc(sizeof(*world)))==NULL)
		{fputs("### no memory for world allocation",stderr); goto cancel;}
	memset(world,0,sizeof(*world));
	if ((players=malloc(sizeof(*players)))==NULL)
		{fputs("### no memory for players",stderr); goto cancel;}
	memset(players,0,sizeof(*players));
	world->endpoint_count = world->triangle_count = 0;
	world->player_count = 1, world->local_player = 0;

	/* default player settings */
	players->sx = players->sy = players->sz = 8192, players->sd = 1024;
	players->terminal_velocity = DEFAULT_TERMINAL_VELOCITY;

	players->walking_constants.mxy = players->walking_constants.myz =
		players->walking_constants.mzx = 4294967296.0/240.0;
	players->walking_constants.forward_magnitude = 8192/60;
	players->walking_constants.positional_deceleration = 8192/20;
	players->walking_constants.turning_deceleration = 4294967296.0/90.0;
	players->walking_constants.terminal_turning_magnitude = 4294967296.0/24.0;
	players->walking_constants.terminal_acceleration = 8192/20;

	players->running_constants.mxy = players->running_constants.myz =
		players->running_constants.mzx = 4294967296.0/120.0;
	players->running_constants.forward_magnitude = 8192/30;
	players->running_constants.positional_deceleration = 8192/20;
	players->running_constants.turning_deceleration = 4294967296.0/90.0;
	players->running_constants.terminal_turning_magnitude = 4294967296.0/20.0;
	players->running_constants.terminal_acceleration = 8192/17;

	while(fgets(line,MAXIMUM_LINE_READ_SIZE,input)!=NULL)
	{
		const char	*s = line;
		UInt8		c, t = _no_part, n;

		++line_count;
		/* this parser does not do long lines */
		switch(line[strlen(s)-1])
		{
		case '\n':
		case '\r':
			break;

		default:
			fprintf(stderr,"### line #%u is too long (max. %u)\n", line_count,MAXIMUM_LINE_READ_SIZE); goto cancel;
		}

		/* skip whitespace and comments */
		while(1)
		{
			if ((c=*s++)==0 || (c+=-'\n')==0 || (c+='\n'-'\r')==0 || (c+='\r'-COMMENT_CHAR)==0)
				break; /* null, eol or eol comment (#) */
			if ((c+=COMMENT_CHAR-'\t')==0 || (c+='\t'-' ')==0 || (c+=' '-DBLSPC_CHAR)==0)
				continue;
			--s;
			if (t!=_no_part)
			{
				if (t==_endpoint_part)
				{
					world_endpoint	e;

					if ((n=sscanf(s,"[%d %d %d]:%d", &e.x,&e.y,&e.z,&e.v))!=4)
						if ((n=sscanf(s,"[%d %d %d]", &e.x,&e.y,&e.z))!=3)
							{fprintf(stderr,"### line #%u - invalid endpoint definition (#%u)\n", line_count,n); goto cancel;}
						else /* default to line index */
							e.v = line_count;
					if (find_endpoint(endpoints,world->endpoint_count,e.v)>=0)
						{fprintf(stderr,"### line #%u - endpoint conflict for ID %d\n", line_count,e.v); goto cancel;}
					vpushback(endpoints,world->endpoint_count,e);
					break;
				} else if (t==_triangle_part)
				{
					world_triangle	t;

					if ((n=sscanf(s,"[%d %d %d] [%u %u %u]", &t.e0,&t.e1,&t.e2,
							&t.basic.ca,&t.basic.cr,&t.basic.cg,&t.basic.cb))!=7)
						if ((n=sscanf(s,"[%d %d %d]", &t.e0,&t.e1,&t.e2))!=3)
							{fprintf(stderr,"### line #%u - invalid triangle definition (#%u)\n", line_count,n); goto cancel;}
						else
							t.basic = default_normal;
					vpushback(triangles,world->triangle_count,t);
					break;
				} else if (t==_player_part)
				{
					if ((n=sscanf(s,"[%d %d %d] [%d %d %d] [%u %u %u] %d",
							&players->px,&players->py,&players->pz,
							&players->vx,&players->vy,&players->vz,
							&players->xy,&players->yz,&players->zx,
							&players->terminal_velocity))!=10)
						if ((n=sscanf(s,"[%d %d %d] [%d %d %d] [%u %u %u]",
								&players->px,&players->py,&players->pz,
								&players->vx,&players->vy,&players->vz,
								&players->xy,&players->yz,&players->zx))!=9)
							{fprintf(stderr,"### line #%u - invalid player definition (#%u)\n", line_count,n); goto cancel;}
						else
							players->terminal_velocity = DEFAULT_TERMINAL_VELOCITY;
					break;
				} else if (t==_walking_constants_part)
				{
					if ((n=sscanf(s,"[%d %d %d] %d %d %d %d %d",
							&players->walking_constants.mxy,
							&players->walking_constants.myz,
							&players->walking_constants.mzx,
							&players->walking_constants.forward_magnitude,
							&players->walking_constants.positional_deceleration,
							&players->walking_constants.turning_deceleration,
							&players->walking_constants.terminal_turning_magnitude,
							&players->walking_constants.terminal_acceleration))!=8)
						{fprintf(stderr,"### line #%u - invalid walking constants definition (#%u)\n", line_count,n); goto cancel;}
					break;
				} else if (t==_running_constants_part)
				{
					if ((n=sscanf(s,"[%d %d %d] %d %d %d %d %d",
							&players->running_constants.mxy,
							&players->running_constants.myz,
							&players->running_constants.mzx,
							&players->running_constants.forward_magnitude,
							&players->running_constants.positional_deceleration,
							&players->running_constants.turning_deceleration,
							&players->running_constants.terminal_turning_magnitude,
							&players->running_constants.terminal_acceleration))!=8)
						{fprintf(stderr,"### line #%u - invalid walking constants definition (#%u)\n", line_count,n); goto cancel;}
					break;
				}
			} else /*if ((c=(--s)[1])==0 || (c+=-'\t')==0 || (c+='\t'-' ')==0 || (c+=' '-DBLSPC_CHAR)==0) */
			{
				switch(*s++)
				{
				case 'E': case 'e': t = _endpoint_part; break;
				case 'T': case 't': t = _triangle_part; break;
				case 'P': case 'p': t = _player_part; break;

				default:
					fprintf(stderr,"### line #%u - unknown part identifier %c\n", line_count,*s);
					goto cancel;
				}
			}/* else
				{fputs("### keywords are not supported",stderr); goto cancel;}*/
		}
	}

	if (world==NULL)
		DebugStr("\pparse_world_file: world structure is NULL!");

	world->endpoints = endpoints;
	world->triangles = triangles;
	world->players = players;

	if (!reconnect_triangles(world))
	{
	cancel:
		if (world!=NULL)
			dispose_world(world);
		return NULL;
	}

	return world;
}

static world_struct *parse_parameters(int argc, char **argv)
{
	world_struct	*world = NULL;
	FILE			*f;
	int				c;
	char			**a, *s;

	if ((c=argc)<=0 || (a=argv)==NULL)
		return NULL;
	while(--c!=0)
		if (*(s=*++a)=='-')
		{
			if (*++s=='\0')
				{f = stdin; goto got_file;}
			if (strcmp(s,"loadkeys")==0)
			{
				if (--c==0)
					{fprintf(stderr,"### %s - Expected filepath for 'keys'\n", argv[0]); goto err;}
				if ((f=fopen(s=*++a,"rb"))==NULL)
					{fprintf(stderr,"### %s - Failed to open '%s'\n", argv[0],s);}
				if (fread(&local_keys,sizeof(local_keys),1,f)<=0)
					{fprintf(stderr,"## %s - Warning: not all key data was present\n", argv[0]);}
				continue;
			} else if (strcmp(s,"dumpkeys")==0)
			{
				if (--c==0)
					{fprintf(stderr,"### %s - Expected filepath for 'keys'\n", argv[0]); goto err;}
				if ((f=fopen(s=*++a,"wb"))==NULL)
					{fprintf(stderr,"### %s - Failed to open '%s'\n", argv[0],s);}
				fwrite(&local_keys,sizeof(local_keys),1,f);
				fclose(f);
				continue;
			} else if (strcmp(s,"setkeys")==0)
			{
				set_keys(&local_keys);
				continue;
			}
			fprintf(stderr,"### %s - Unrecognized option '%s'\n", argv[0],s);
		err:
			fprintf(stderr,"# Form: %s [options] file\n", argv[0]);
			fputs(
"#\n"
"# -                 use stdin for input\n"
				,stderr);
		} else
		{
			if ((f=fopen(s,"r+"))==NULL)
				{fprintf(stderr,"### %s - Failed to open '%s'\n", argv[0],s); goto err;}
		got_file:
			if ((world=parse_world_file(f))==NULL)
				break;
		}

	return world;
}

/* ---- tables */

#define PRECISION_ANGLE_COUNT		(1L<<12)
#define FULL_TO_PREC_ANGLE(x)		((x)>>32-12)

sfract				*sine_table; /* 12 bit angle precision */
sfract				*cosine_table;

/* 30 bit fraction short sine table for the sine/cosine tables */
/* reordered: +[0,1024], +[1023,1], -[0,1024], -[1023,1], +[0,1023] */
static const sfract		short_sine_table[1025] =
{
	0, 1647099, 3294193, 4941281,
	6588356, 8235416, 9882456, 11529474,
	13176464, 14823423, 16470347, 18117233,
	19764076, 21410872, 23057618, 24704310,
	26350943, 27997515, 29644021, 31290457,
	32936819, 34583104, 36229307, 37875426,
	39521455, 41167391, 42813230, 44458968,
	46104602, 47750128, 49395541, 51040837,
	52686014, 54331067, 55975992, 57620785,
	59265442, 60909960, 62554335, 64198563,
	65842639, 67486561, 69130324, 70773924,
	72417357, 74060620, 75703709, 77346620,
	78989349, 80631892, 82274245, 83916404,
	85558366, 87200127, 88841683, 90483029,
	92124163, 93765079, 95405776, 97046247,
	98686491, 100326502, 101966277, 103605812,
	105245103, 106884147, 108522939, 110161476,
	111799753, 113437768, 115075515, 116712992,
	118350194, 119987118, 121623759, 123260114,
	124896179, 126531950, 128167423, 129802595,
	131437462, 133072019, 134706263, 136340190,
	137973796, 139607077, 141240030, 142872651,
	144504935, 146136880, 147768480, 149399733,
	151030634, 152661180, 154291367, 155921191,
	157550647, 159179733, 160808445, 162436778,
	164064728, 165692293, 167319468, 168946249,
	170572633, 172198615, 173824192, 175449360,
	177074115, 178698453, 180322371, 181945865,
	183568930, 185191564, 186813762, 188435520,
	190056834, 191677702, 193298119, 194918080,
	196537583, 198156624, 199775198, 201393302,
	203010932, 204628085, 206244756, 207860942,
	209476638, 211091842, 212706549, 214320755,
	215934457, 217547651, 219160334, 220772500,
	222384147, 223995270, 225605867, 227215933,
	228825464, 230434456, 232042906, 233650811,
	235258165, 236864966, 238471210, 240076892,
	241682010, 243286558, 244890535, 246493935,
	248096755, 249698991, 251300640, 252901697,
	254502159, 256102022, 257701283, 259299937,
	260897982, 262495412, 264092224, 265688415,
	267283981, 268878918, 270473223, 272066891,
	273659918, 275252302, 276844038, 278435122,
	280025552, 281615322, 283204430, 284792871,
	286380643, 287967740, 289554160, 291139898,
	292724951, 294309316, 295892988, 297475964,
	299058239, 300639811, 302220676, 303800829,
	305380268, 306958988, 308536985, 310114257,
	311690799, 313266607, 314841679, 316416009,
	317989595, 319562433, 321134518, 322705848,
	324276419, 325846226, 327415267, 328983538,
	330551034, 332117752, 333683689, 335248841,
	336813204, 338376774, 339939549, 341501523,
	343062693, 344623057, 346182609, 347741347,
	349299266, 350856364, 352412636, 353968079,
	355522689, 357076462, 358629395, 360181484,
	361732726, 363283116, 364832652, 366381329,
	367929144, 369476093, 371022173, 372567379,
	374111709, 375655159, 377197725, 378739403,
	380280190, 381820082, 383359076, 384897167,
	386434353, 387970630, 389505993, 391040440,
	392573967, 394106570, 395638246, 397168991,
	398698801, 400227673, 401755603, 403282588,
	404808624, 406333708, 407857835, 409381002,
	410903207, 412424444, 413944711, 415464004,
	416982319, 418499653, 420016002, 421531363,
	423045732, 424559105, 426071480, 427582852,
	429093217, 430602573, 432110916, 433618242,
	435124548, 436629829, 438134084, 439637307,
	441139496, 442640647, 444140756, 445639820,
	447137835, 448634799, 450130706, 451625555,
	453119340, 454612060, 456103710, 457594286,
	459083786, 460572205, 462059541, 463545789,
	465030947, 466515010, 467997976, 469479840,
	470960600, 472440251, 473918791, 475396216,
	476872522, 478347705, 479821764, 481294693,
	482766489, 484237150, 485706671, 487175049,
	488642281, 490108363, 491573292, 493037064,
	494499676, 495961124, 497421405, 498880516,
	500338453, 501795212, 503250791, 504705185,
	506158392, 507610408, 509061229, 510510853,
	511959275, 513406493, 514852502, 516297300,
	517740883, 519183248, 520624391, 522064309,
	523502998, 524940456, 526376678, 527811662,
	529245404, 530677900, 532109148, 533539144,
	534967884, 536395365, 537821584, 539246538,
	540670223, 542092635, 543513772, 544933630,
	546352205, 547769495, 549185496, 550600205,
	552013618, 553425732, 554836544, 556246051,
	557654248, 559061133, 560466703, 561870954,
	563273883, 564675486, 566075761, 567474703,
	568872310, 570268579, 571663506, 573057087,
	574449320, 575840202, 577229728, 578617896,
	580004702, 581390144, 582774218, 584156920,
	585538248, 586918198, 588296766, 589673951,
	591049748, 592424154, 593797166, 595168781,
	596538995, 597907806, 599275210, 600641203,
	602005783, 603368947, 604730691, 606091012,
	607449906, 608807372, 610163404, 611518001,
	612871159, 614222875, 615573145, 616921967,
	618269338, 619615253, 620959711, 622302707,
	623644239, 624984303, 626322897, 627660017,
	628995660, 630329823, 631662503, 632993696,
	634323400, 635651611, 636978327, 638303543,
	639627258, 640949467, 642270169, 643589359,
	644907034, 646223192, 647537830, 648850943,
	650162530, 651472587, 652781111, 654088099,
	655393548, 656697454, 657999816, 659300629,
	660599890, 661897597, 663193747, 664488336,
	665781362, 667072820, 668362709, 669651026,
	670937767, 672222928, 673506508, 674788504,
	676068911, 677347728, 678624950, 679900576,
	681174602, 682447025, 683717842, 684987051,
	686254647, 687520629, 688784993, 690047736,
	691308855, 692568348, 693826211, 695082441,
	696337036, 697589992, 698841307, 700090977,
	701339000, 702585372, 703830092, 705073155,
	706314559, 707554301, 708792378, 710028787,
	711263525, 712496590, 713727978, 714957687,
	716185713, 717412054, 718636707, 719859669,
	721080937, 722300508, 723518380, 724734549,
	725949013, 727161768, 728372813, 729582143,
	730789757, 731995651, 733199822, 734402269,
	735602987, 736801974, 737999228, 739194745,
	740388522, 741580558, 742770848, 743959390,
	745146182, 746331221, 747514503, 748696026,
	749875788, 751053785, 752230015, 753404474,
	754577161, 755748072, 756917205, 758084557,
	759250125, 760413906, 761575898, 762736098,
	763894504, 765051111, 766205919, 767358923,
	768510122, 769659512, 770807092, 771952857,
	773096806, 774238936, 775379244, 776517728,
	777654384, 778789210, 779922204, 781053363,
	782182683, 783310163, 784435800, 785559591,
	786681534, 787801625, 788919863, 790036244,
	791150767, 792263427, 793374223, 794483153,
	795590213, 796695401, 797798714, 798900150,
	799999706, 801097379, 802193167, 803287068,
	804379079, 805469196, 806557419, 807643743,
	808728167, 809810688, 810891304, 811970011,
	813046808, 814121692, 815194659, 816265709,
	817334838, 818402043, 819467323, 820530675,
	821592095, 822651583, 823709135, 824764748,
	825818421, 826870150, 827919934, 828967769,
	830013654, 831057586, 832099562, 833139580,
	834177638, 835213733, 836247863, 837280024,
	838310216, 839338435, 840364679, 841388945,
	842411232, 843431536, 844449856, 845466188,
	846480531, 847492882, 848503239, 849511600,
	850517961, 851522321, 852524677, 853525028,
	854523370, 855519701, 856514019, 857506321,
	858496606, 859484870, 860471112, 861455330,
	862437520, 863417681, 864395810, 865371905,
	866345964, 867317984, 868287963, 869255900,
	870221790, 871185633, 872147426, 873107167,
	874064853, 875020483, 875974054, 876925563,
	877875009, 878822389, 879767701, 880710943,
	881652112, 882591207, 883528225, 884463164,
	885396022, 886326796, 887255485, 888182086,
	889106597, 890029016, 890949341, 891867569,
	892783698, 893697727, 894609652, 895519473,
	896427186, 897332790, 898236282, 899137661,
	900036924, 900934069, 901829095, 902721998,
	903612776, 904501429, 905387953, 906272347,
	907154608, 908034735, 908912725, 909788576,
	910662286, 911533853, 912403276, 913270551,
	914135678, 914998653, 915859476, 916718143,
	917574653, 918429004, 919281194, 920131221,
	920979082, 921824777, 922668302, 923509656,
	924348837, 925185843, 926020672, 926853322,
	927683790, 928512076, 929338177, 930162092,
	930983817, 931803352, 932620694, 933435842,
	934248793, 935059546, 935868098, 936674448,
	937478595, 938280535, 939080267, 939877790,
	940673101, 941466198, 942257081, 943045745,
	943832191, 944616416, 945398418, 946178196,
	946955747, 947731070, 948504163, 949275023,
	950043650, 950810042, 951574196, 952336111,
	953095785, 953853216, 954608403, 955361344,
	956112036, 956860479, 957606670, 958350608,
	959092290, 959831716, 960568883, 961303790,
	962036435, 962766816, 963494932, 964220780,
	964944360, 965665669, 966384706, 967101468,
	967815955, 968528165, 969238095, 969945745,
	970651112, 971354196, 972054994, 972753504,
	973449725, 974143656, 974835295, 975524639,
	976211688, 976896441, 977578894, 978259047,
	978936898, 979612445, 980285688, 980956623,
	981625251, 982291568, 982955574, 983617267,
	984276646, 984933708, 985588453, 986240879,
	986890984, 987538766, 988184225, 988827359,
	989468165, 990106644, 990742793, 991376610,
	992008094, 992637245, 993264059, 993888536,
	994510675, 995130473, 995747930, 996363043,
	996975812, 997586236, 998194311, 998800038,
	999403415, 1000004439, 1000603111, 1001199428,
	1001793390, 1002384994, 1002974239, 1003561124,
	1004145648, 1004727809, 1005307605, 1005885036,
	1006460100, 1007032796, 1007603122, 1008171077,
	1008736660, 1009299870, 1009860704, 1010419162,
	1010975242, 1011528943, 1012080264, 1012629204,
	1013175761, 1013719934, 1014261721, 1014801122,
	1015338134, 1015872758, 1016404991, 1016934832,
	1017462281, 1017987335, 1018509994, 1019030256,
	1019548121, 1020063586, 1020576651, 1021087314,
	1021595575, 1022101432, 1022604883, 1023105929,
	1023604567, 1024100796, 1024594615, 1025086024,
	1025575020, 1026061603, 1026545772, 1027027525,
	1027506862, 1027983780, 1028458280, 1028930359,
	1029400018, 1029867254, 1030332067, 1030794455,
	1031254418, 1031711954, 1032167062, 1032619742,
	1033069992, 1033517810, 1033963197, 1034406151,
	1034846671, 1035284755, 1035720404, 1036153615,
	1036584389, 1037012723, 1037438617, 1037862069,
	1038283080, 1038701647, 1039117770, 1039531448,
	1039942680, 1040351465, 1040757802, 1041161689,
	1041563127, 1041962114, 1042358649, 1042752731,
	1043144360, 1043533534, 1043920252, 1044304514,
	1044686319, 1045065665, 1045442553, 1045816980,
	1046188946, 1046558451, 1046925492, 1047290071,
	1047652185, 1048011834, 1048369016, 1048723732,
	1049075980, 1049425759, 1049773069, 1050117909,
	1050460278, 1050800175, 1051137599, 1051472550,
	1051805027, 1052135029, 1052462555, 1052787604,
	1053110176, 1053430270, 1053747885, 1054063021,
	1054375676, 1054685850, 1054993543, 1055298753,
	1055601479, 1055901722, 1056199480, 1056494753,
	1056787540, 1057077840, 1057365653, 1057650977,
	1057933813, 1058214159, 1058492016, 1058767381,
	1059040255, 1059310638, 1059578527, 1059843923,
	1060106826, 1060367233, 1060625146, 1060880563,
	1061133483, 1061383907, 1061631833, 1061877261,
	1062120190, 1062360620, 1062598550, 1062833980,
	1063066909, 1063297336, 1063525261, 1063750684,
	1063973603, 1064194019, 1064411931, 1064627338,
	1064840240, 1065050636, 1065258526, 1065463909,
	1065666786, 1065867154, 1066065015, 1066260367,
	1066453210, 1066643544, 1066831367, 1067016680,
	1067199483, 1067379774, 1067557554, 1067732821,
	1067905576, 1068075818, 1068243547, 1068408763,
	1068571464, 1068731650, 1068889322, 1069044479,
	1069197120, 1069347245, 1069494854, 1069639946,
	1069782521, 1069922579, 1070060120, 1070195142,
	1070327646, 1070457632, 1070585099, 1070710046,
	1070832474, 1070952382, 1071069770, 1071184638,
	1071296985, 1071406812, 1071514117, 1071618901,
	1071721163, 1071820903, 1071918122, 1072012818,
	1072104991, 1072194642, 1072281769, 1072366374,
	1072448455, 1072528012, 1072605046, 1072679556,
	1072751542, 1072821003, 1072887940, 1072952352,
	1073014240, 1073073603, 1073130440, 1073184753,
	1073236540, 1073285802, 1073332538, 1073376748,
	1073418433, 1073457592, 1073494225, 1073528332,
	1073559913, 1073588967, 1073615496, 1073639498,
	1073660973, 1073679922, 1073696345, 1073710241,
	1073721611, 1073730454, 1073736771, 1073740561,
	1073741824,
};

static Boolean initialize_trig_tables(void)
{
	int				i;

	/* the sine table is expanded to cover the cosine table at quarter phase */
	if ((sine_table=malloc(5*PRECISION_ANGLE_COUNT))==NULL) /* (5/4)*sizeof(sfract) */
		return false;

	memcpy(sine_table,short_sine_table,sizeof(short_sine_table));
	memcpy(cosine_table=sine_table+PRECISION_ANGLE_COUNT,sine_table,PRECISION_ANGLE_COUNT); /* PRECISION_ANGLE_COUNT*sizeof(sfract)/4 */

	/* second quarter is really the first in reverse */
	for(i=1025; i<=2047; ++i)
		sine_table[i] = sine_table[2048-i];
	/* second half is really the negative of the first */
	for(i=2048; i<=4095; ++i)
		sine_table[i] = -sine_table[i-2048];

	return true;
}

static void free_trig_tables(void)
{
	if (sine_table!=NULL)
		{free(sine_table); sine_table = NULL;}
}

/* ---- world calculations */

#if TARGET_CPU_68K
# pragma parameter sqsum2(__A0,__D1,__D2)
extern void sqsum2(UInt64*, SInt32, SInt32) = {0x4C01, 0x1C00, 0x2241, 0x4C02, 0x2C01,
		0xD489, 0xD380, 0x20C1, 0x2082};
# pragma parameter sqsum3(__A0,__D0,__D1,__D2)
extern void sqsum3(UInt64*, SInt32, SInt32, SInt32) = {0x2F00, 0x4C01, 0x1C00, 0x2241,
		0x4C02, 0x2C01, 0xD489, 0x2242, 0xD181, 0x241F, 0x4C02, 0x2C01, 0xD489,
		0xD380, 0x20C1, 0x2082};
#else
	_inline static UInt64 sqsum2(SInt32 a, SInt32 b)
	{
		return (long long)a*a+(long long)b*b;
	}
	_inline static UInt64 sqsum3(SInt32 a, SInt32 b, SInt32 c)
	{
		return (long long)a*a+(long long)b*b+(long long)c*c;
	}
#endif

UInt32 ihypot2(SInt32 a, SInt32 b)
{
#if TARGET_CPU_68K
	UInt64			_t;

	sqsum2(&_t,a,b);
	return isqrt64(_t);
#else
	return isqrt64(sqsum2(a,b));
#endif
}

UInt32 ihypot3(SInt32 a, SInt32 b, SInt32 c)
{
#if TARGET_CPU_68K
	UInt64			_t;

	sqsum3(&_t,a,b,c);
	return isqrt64(_t);
#else
	return isqrt64(sqsum3(a,b,c));
#endif
}

/* ---- main constants */

enum
{ /* action flags */
	_move_forward,
	_move_backward,
	_sidestep_left,
	_sidestep_right,
	_turn_left, /* zx */
	_turn_right,
	_turn_up, /* yz */
	_turn_down,
	_y_recenter, /* recenter xy and yz */
	_run_dont_walk,
	_roll_left, /* xy */
	_roll_right,
	_zoom_in,
	_zoom_out,
	
	NUMBER_OF_ACTION_FLAGS /* <=32 */
};

/* ---- main globals */

QDGlobals			qd;

WindowRef			world_window = NULL;
CGrafPtr			world_port = NULL;

Rect				port_rectangle, dest_rectangle;

/* ---- main prototypes */

void initialize_world(world_struct *world);
void run_world(world_struct *world);

/* ---- main code */

static void initialize_program(void)
{
/*	MoreMasters();
	MoreMasters();
	MoreMasters();
	MaxApplZone();*/

	InitGraf((Ptr)&qd.thePort);
	FlushEvents(everyEvent,0);
}

static Boolean initialize_screen(void)
{
	QDErr			err;
	int				screen_width, screen_height;

	dest_rectangle = (**GetMainDevice()).gdRect;
	if ((screen_width=RECT_WIDTH(&dest_rectangle))<WORLD_WIDTH
			|| (screen_height=RECT_HEIGHT(&dest_rectangle))<WORLD_HEIGHT)
	{
		fprintf(stderr,"### Screen is too small (minimum WxH: %ux%u).\n", WORLD_WIDTH,WORLD_HEIGHT);
		return false;
	}

	port_rectangle.top = port_rectangle.left = 0;
	port_rectangle.bottom = WORLD_HEIGHT, port_rectangle.right = WORLD_WIDTH;
	adjust_rectangle(&dest_rectangle,&port_rectangle,&dest_rectangle,_adjust_rectangle_center);
	OFFSET_RECTANGLE(&dest_rectangle,-(dest_rectangle.left&7),-(dest_rectangle.top&7));

	if ((window=NewCWindow(NewPtr(sizeof(CWindowRecord)),&dest_rectangle,empty,true,0,
			(WindowRef)-1,false,0))==NULL)
		{DebugStr("\initialize_screen: NewCWindow failed"); return false;}
	dest_rectangle = port_rectangle;
	if ((err=NewGWorld(&world_port,BIT_DEPTH,&port_rectangle,NULL,NULL,0))!=0)
		{fprintf(stderr,"### NewGWorld: error #%d", err); return false;}

	return true;
}

const key_settings	default_keys =
{
	{ 37,  0,  0,'i'}, /* move forward */
	{ 47,  0,  0,'k'}, /* move backward */

	{ 39,  0,  0,'u'}, /* sidestep left */
	{ 24,  0,  0,'o'}, /* sidestep right */

	{ 33,  0,  0,'j'}, /* turn left */
	{ 32,  0,  0,'l'}, /* turn right */
	{  3,  0,  0,'h'}, /* turn up */
	{ 42,  0,  0,'n'}, /* turn down */

	{ 44,  0,  0,','}, /* y recenter */

	{ 63,  0,'S','h'}, /* run dont walk */

	{ 41,  0,  0,'m'}, /* roll left */
	{ 40,  0,  0,'.'}, /* roll right */

	{ 46,  0,  0,';'}, /* zoom in */
	{ 43,  0,  0,'/'}, /* zoom out */
};

void main(int argc, char **argv)
{
	world_struct	*world;

	initialize_program();
	memcpy(local_keys,default_keys,sizeof(local_keys));
	Debugger();
	if ((world=parse_parameters(argc,argv))==NULL)
		exit(1);

	initialize_world(world);
	if (!initialize_trig_tables()
			|| !initialize_screen())
		dispose_world(world);
	else
		run_world(world);

	FlushEvents(everyEvent,0);

	if (world_port!=NULL)
		DisposeGWorld(world_port);

	free_trig_tables();

	exit(0);
}

#define KEY_NAME_INDENT				8
#define KEY_CHAR_LEFT				128
#define KEY_LINE_HEIGHT				16
#define KEY_Y_FUDGE					2

void set_keys(key_settings *keys)
{
	typedef UInt8		Str7[8];

	static const Str7	key_chars[128] =
	{ /* note that these have been reordered high-order bitwise */
		"\px",
		"\pz",
		"\pg",
		"\ph",
		"\pf",
		"\pd",
		"\ps",
		"\pa",
		"\pr", /* 8- */
		"\pe",
		"\pw",
		"\pq",
		"\pb",
		"\p0x0D",
		"\pv",
		"\pc",
		"\p5", /* 16- */
		"\p6",
		"\p4",
		"\p3",
		"\p2",
		"\p1",
		"\pt",
		"\py",
		"\po", /* 24- */
		"\p]",
		"\p0",
		"\p8",
		"\p-",
		"\p7",
		"\p9",
		"\p=",
		"\p\047", /* 32- */
		"\pj",
		"\pl (L)",
		"\pReturn",
		"\pp",
		"\pi",
		"\p[",
		"\pu",
		"\p.", /* 40- */
		"\pm",
		"\pn",
		"\p/",
		"\p,",
		"\p\134",
		"\p;",
		"\pk",
		"\pCmd", /* 48- */
		"\p0x31",
		"\pEsc",
		"\p0x33",
		"\pBackspc",
		"\p`",
		"\pSpace",
		"\pTab",
		"\p0x38", /* 56- */
		"\p0x39",
		"\p0x3A",
		"\p0x3B",
		"\pContrl",
		"\pOptn",
		"\pCapsLck",
		"\pShift",
		"\pClear", /* 64- */
		"\p0x41",
		"\pKeypd +",
		"\p0x43",
		"\pKeypd *",
		"\p0x45",
		"\pKeypd .",
		"\p0x47",
		"\p0x48", /* 72- */
		"\pKeypd -",
		"\p0x4A",
		"\pEnter",
		"\pKeypd /",
		"\p0x4D",
		"\p0x4E",
		"\p0x4F",
		"\pKeypd 5", /* 80- */
		"\pKeypd 4",
		"\pKeypd 3",
		"\pKeypd 2",
		"\pKeypd 1",
		"\pKeypd 0",
		"\pKeypd =",
		"\p0x57",
		"\p0x58", /* 88- */
		"\p0x59",
		"\p0x5A",
		"\pKeypd 9",
		"\pKeypd 8",
		"\p0x5D",
		"\pKeypd 7",
		"\pKeypd 6",
		"\pF11", /* 96- */
		"\p0x61",
		"\pF9",
		"\pF8",
		"\pF3",
		"\pF7",
		"\pF6",
		"\pF5",
		"\pF12", /* 104- */
		"\p0x69",
		"\pF10",
		"\p0x6B",
		"\pF14",
		"\p0x6D",
		"\pF13",
		"\p0x6F",
		"\pEnd", /* 112- */
		"\pF4",
		"\pFwd Del",
		"\pPage Up",
		"\pHome",
		"\pHelp",
		"\pF15",
		"\p0x77",
		"\pPower?!", /* 120- */
		"\pAr Up",
		"\pAr Down",
		"\pAr Rgt",
		"\pAr Left",
		"\pF1",
		"\pPage Dn",
		"\pF2",
	};

	static const Str15	action_strings[NUMBER_OF_ACTION_FLAGS] =
	{
		"\pMove Forward",
		"\pMove Backward",
		"\pSidestep Left",
		"\pSidestep Right",
		"\pTurn Left",
		"\pTurn Right",
		"\pTurn Down",
		"\pTurn Up",
		"\pY Recenter",
		"\pRun",
		"\pRoll Left",
		"\pRoll Right",
		"\pZoom In",
		"\pZoom Out",
	};

	key_settings	keys_copy;
	Rect			bounds;
	RGBColor		color;

	GrafPtr			old_port;
	UInt8			*str;
	int				x = NUMBER_OF_ACTION_FLAGS;

	GetPort(&old_port);

	memcpy(keys_copy,keys,sizeof(keys_copy));

	SetPortWindowPort(world_window);

	MovePortTo(32,64);
	bounds.top = 0, bounds.left = 0, bounds.bottom = 320, bounds.right = 224;
	EraseRect(&bounds);

	TextFont(0);
	TextSize(12);
	TextMode(blend);
	color.red = color.green = color.blue = 32767;
	OpColor(&color);

	str = "\p====  Set Keys...  ====";
	MoveTo((bounds.right+bounds.left-StringWidth(str))>>1,bounds.top-15);
	DrawString(str);
	TextMode(srcOr);
	str = "\pSet Keys...";
	MoveTo((bounds.right+bounds.left-StringWidth(str))>>1,bounds.top-16);
	DrawString(str);

	/* render whole */
	for(x=0; x<NUMBER_OF_ACTION_FLAGS; ++x)
	{
		MoveTo(KEY_NAME_INDENT,KEY_NAME_INDENT+(x+1)*KEY_LINE_HEIGHT);
		DrawString(action_strings[x]);
		MoveTo(KEY_CHAR_LEFT,KEY_NAME_INDENT+(x+1)*KEY_LINE_HEIGHT);
		if ((keys_copy[x].char_code|keys_copy[x].char_index)==0)
			DrawString("\p<unset>");
		else
			DrawString(key_chars[keys_copy[x].map_index]);
	}

	MoveTo(KEY_NAME_INDENT*2,KEY_NAME_INDENT+(x+1)*KEY_LINE_HEIGHT);
	DrawString("\pCancel");
	MoveTo(KEY_NAME_INDENT*2,KEY_NAME_INDENT+(x+2)*KEY_LINE_HEIGHT);
	DrawString("\pOkay");

	x = 0;
	MoveTo(KEY_CHAR_LEFT-1,KEY_NAME_INDENT+KEY_Y_FUDGE+x*KEY_LINE_HEIGHT);
	Line(0,KEY_LINE_HEIGHT+KEY_Y_FUDGE);
	while(1)
	{
		EventRecord		event;

		if (GetNextEvent(mUpMask|keyDownMask|autoKeyMask,&event))
			if (event.what==keyDown || event.what==autoKey)
			{
				KeyMap		keymap, oldmap;

				key_setting		*key;

				GetKeys(oldmap);
				switch(event.message&charCodeMask)
				{
				case kUpArrowCharCode:
				case kLeftArrowCharCode:
					PenMode(srcBic);
					MoveTo(KEY_CHAR_LEFT-1,KEY_NAME_INDENT+KEY_Y_FUDGE+x*KEY_LINE_HEIGHT);
					Line(0,KEY_LINE_HEIGHT+KEY_Y_FUDGE);
					PenMode(srcCopy);
					if (--x<0)
						x += NUMBER_OF_ACTION_FLAGS+1;
					MoveTo(KEY_CHAR_LEFT-1,KEY_NAME_INDENT+KEY_Y_FUDGE+x*KEY_LINE_HEIGHT);
					Line(0,KEY_LINE_HEIGHT+KEY_Y_FUDGE);
					break;
				case kDownArrowCharCode:
				case kRightArrowCharCode:
					PenMode(srcBic);
					MoveTo(KEY_CHAR_LEFT-1,KEY_NAME_INDENT+KEY_Y_FUDGE+x*KEY_LINE_HEIGHT);
					Line(0,KEY_LINE_HEIGHT+KEY_Y_FUDGE);
					PenMode(srcCopy);
					if (++x>NUMBER_OF_ACTION_FLAGS+1)
						x -= NUMBER_OF_ACTION_FLAGS+1;
					MoveTo(KEY_CHAR_LEFT-1,KEY_NAME_INDENT+KEY_Y_FUDGE+x*KEY_LINE_HEIGHT);
					Line(0,KEY_LINE_HEIGHT+KEY_Y_FUDGE);
					break;
				case kReturnCharCode:
				case kEnterCharCode:
					if (x>=NUMBER_OF_ACTION_FLAGS)
					{
						if (x>NUMBER_OF_ACTION_FLAGS)
							memcpy(keys,keys_copy,sizeof(*keys));
				case kEscapeCharCode:
						goto end;
					}

					/* wait for depress first */
					do{GetKeys(keymap);} while(memcmp(keymap,oldmap,sizeof(keymap))==0);

					bounds.left = KEY_CHAR_LEFT, bounds.right = 224-KEY_NAME_INDENT;
					bounds.bottom = KEY_LINE_HEIGHT+KEY_Y_FUDGE+(bounds.top=KEY_NAME_INDENT+KEY_Y_FUDGE+x*KEY_LINE_HEIGHT);
					PenMode(srcXor);
					PaintRect(&bounds);
					PenMode(srcCopy);

					GetKeys(oldmap);
					oldmap[3] &= ~0x80; /* clear Power key */
					key = x+keys_copy;
					while(1)
					{
						GetKeys(keymap);
						keymap[3] &= ~0x80; /* clear Power key */
						if ((keymap[0]^=oldmap[0])!=0)
							{key->map_index = LZCount32(keymap[0]); break;}
						if ((keymap[1]^=oldmap[1])!=0)
							{key->map_index = 32+LZCount32(keymap[1]); break;}
						if ((keymap[2]^=oldmap[2])!=0)
							{key->map_index = 64+LZCount32(keymap[2]); break;}
						if ((keymap[3]^=oldmap[3])!=0)
							{key->map_index = 96+LZCount32(keymap[3]); break;}
					}

					EraseRect(&bounds);
					MoveTo(KEY_CHAR_LEFT,KEY_NAME_INDENT+(x+1)*KEY_LINE_HEIGHT);
					DrawString(key_chars[key->map_index]);
					FlushEvents(everyEvent,0); /* clear the key hit */

					/* update previous and next in case of overlap */
					if (x>0)
					{
						MoveTo(KEY_CHAR_LEFT,KEY_NAME_INDENT+x*KEY_LINE_HEIGHT);
						if ((keys_copy[x-1].char_code|keys_copy[x-1].char_index)==0)
							DrawString("\p<unset>");
						else
							DrawString(key_chars[keys_copy[x-1].map_index]);
					}
					if (x<NUMBER_OF_ACTION_FLAGS-1)
					{
						MoveTo(KEY_CHAR_LEFT,KEY_NAME_INDENT+(x+2)*KEY_LINE_HEIGHT);
						if ((keys_copy[x+1].char_code|keys_copy[x+1].char_index)==0)
							DrawString("\p<unset>");
						else
							DrawString(key_chars[keys_copy[x+1].map_index]);
					}

					/* crude, but alternatively effective versus using system b.s. */
					str = key_chars[key->map_index];
					switch(str[1])
					{
					case 'A': /* arrow keys */
						key->char_code = 'A', key->char_index = str[4];
						break;
					case 'B': /* backspace */
						key->char_code = 'D', key->char_index = 8;
						break;
					case 'F': /* function keys, foward delete */
						key->char_code = 'F';
						if (str[0]==2)
							key->char_index = str[2]-'0';
						else /* forward delete will come out 0xFA */
							key->char_index = (unsigned short_int)str[2]*10+str[3]+-('0'*10+'0');
						break;
					case 'C': /* commanD, caps locK, controL, cleaR */
					case 'E': /* escapE, enteR, enD */
					case 'H': /* help, home */
					case 'K': /* keypad */
					case 'O': /* option */
					case 'P': /* page uP/dowN, power (not applied) */
					case 'R': /* return */
						key->char_code = str[1], key->char_index = str[str[0]];
						break;
					case 'S': /* space, shift */
						if (str[2]!='p')
							key->char_code = 'S', key->char_index = str[2];
						else
							key->char_code = 0, key->char_index = ' ';
						break;
					case 'T': /* tab */
						key->char_code = 0;
						key->char_index = 9;
						break;
					case '0':
						if (str[0]>1)
						{
							key->char_code = '?';
							key->char_index = ((str[3]<<4)-~(str[3]&16)*7)|((str[4]-~((str[4]>>4)&1)*7)&15);
							break;
						}
					default:
						key->char_code = 0;
						key->char_index = str[str[0]];
					}
					break;
				}
			} else
				break;
	}
end:
	/*DisposeWindow(window);*/
	MovePortTo(0,0);
	FlushEvents(everyEvent,0);

	SetPort(old_port);
}

static UInt32 update_keys(const key_setting *keys)
{
	KeyMap			keymap;
	UInt32			action_flags = 0;
	int				x = NUMBER_OF_ACTION_FLAGS, v = 1; /* x = sizeof(*keys)/sizeof(keys[0]) */

	GetKeys(keymap);
	do
	{
		if (keymap[keys->map_index>>5]&(1L<<((~keys->map_index)&31)))
			action_flags += v;
		++keys, v += v;
	} while(--x);

	return 0;
}

static void move_player(world_player *player, UInt32 action_flags)
{
	player->old_action_flags = player->new_action_flags;
	player->new_action_flags = action_flags;

	/* update velocities and turning magnitudes */
	player->vx += player->ax, player->vy += player->ay, player->vz += player->az;
	player->xy += player->vxy, player->yz += player->vyz, player->zx += player->vzx;

	/* clip angular velocity */
	if ((player->angular_velocity=ihypot3(player->vx,player->vy,player->vz))>player->terminal_velocity)
	{ /* clip velocity by ratio of terminal to total */
		sfract		ratio = NFix30Div(player->angular_velocity,player->terminal_velocity);

		player->vx = NFix30Mul(player->vx,ratio);
		player->vy = NFix30Mul(player->vy,ratio);
		player->vz = NFix30Mul(player->vz,ratio);
	}
	player->px += player->vx, player->py += player->vy, player->pz += player->vz;
}

static void update_player(world_player *player, UInt32 action_flags)
{
	/* table for converting forward/backward/sidestepping actions into a partial
		quadrant (+1 so 0 means to decelerate) -- also applies cancellation */
	static UInt8	movement_quadrants[16] = {0,1,5,0, 3,2,4,3, 7,8,6,7, 0,1,5,0};

	acceleration_constants	*constants = (action_flags&_run_dont_walk) ?
		&player->running_constants:&player->walking_constants;

	UInt32			departure_angle;

	/* update accelerations according to the action flags */
	switch(action_flags&(_turn_left|_turn_right))
	{
	case _turn_left: player->vzx += constants->mzx; break;
	case _turn_right: player->vzx -= constants->mzx; break;

	default: /* none or both: decelerate */
		if (player->vzx>0)
		{
			if ((player->vzx-=constants->turning_deceleration)<0)
				player->vzx = 0;
		} else if ((player->vzx+=constants->turning_deceleration)>0)
			player->vzx = 0;
	}

	if (action_flags&_y_recenter)
	{ /* be sure to clip magnitudes */
		if ((player->yz&(long)0x80000000)==0) /* looking up */
		{
			if ((unsigned)-(player->vyz-=constants->myz)>player->yz)
				player->vyz = -player->yz;
		} else if ((unsigned)-(player->vyz+=constants->myz)<player->yz)
			player->vyz = -constants->myz;
		if ((player->xy&(long)0x80000000)==0) /* looking sideways left */
		{
			if ((unsigned)-(player->vxy-=constants->mxy)>player->xy)
				player->vxy = -player->xy;
		} else if ((unsigned)-(player->vxy+=constants->mxy)<player->xy)
			player->vxy = -constants->mxy;
	} else
		switch(action_flags&(_turn_up|_turn_down))
		{
		case _turn_up: player->vyz += constants->myz; break;
		case _turn_down: player->vyz -= constants->myz; break;
	
		default: /* none or both: decelerate */
			if (player->vyz>0)
			{
				if ((player->vyz-=constants->turning_deceleration)<0)
					player->vyz = 0;
			} else if ((player->vyz+=constants->turning_deceleration)>0)
				player->vyz = 0;
		}

	if ((departure_angle=movement_quadrants[action_flags&(_move_forward|_move_backward|_sidestep_left|_sidestep_right)])==0)
	{ /* decelerate departure */
		if (player->angular_acceleration>0)
		{
			if (-(player->angular_acceleration-=constants->positional_deceleration)>player->angular_velocity)
				player->angular_acceleration = -player->angular_velocity;
		} else if ((player->angular_acceleration+=constants->positional_deceleration)>player->angular_velocity)
			player->angular_acceleration = player->angular_velocity;
	} else
	{ /* convert the adjusted quadrant to the angle of departure */
		departure_angle = player->zx+(departure_angle-1<<32-3);
		departure_angle = FULL_TO_PREC_ANGLE(departure_angle);

		if ((action_flags&(_sidestep_left|_sidestep_right))!=0)
			player->angular_acceleration = constants->forward_magnitude>>1;
		else if ((action_flags&_move_backward)!=0)
			player->angular_acceleration = constants->forward_magnitude-(constants->forward_magnitude>>2);
		else
			player->angular_acceleration = constants->forward_magnitude;
	}

	/* clip accelerations */
	if (player->vxy>constants->terminal_turning_magnitude)
		player->vxy = constants->terminal_turning_magnitude;
	else if (player->vxy<-constants->terminal_turning_magnitude)
		player->vxy = -constants->terminal_turning_magnitude;
	if (player->vyz>constants->terminal_turning_magnitude)
		player->vyz = constants->terminal_turning_magnitude;
	else if (player->vyz<-constants->terminal_turning_magnitude)
		player->vyz = -constants->terminal_turning_magnitude;
	if (player->vzx>constants->terminal_turning_magnitude)
		player->vzx = constants->terminal_turning_magnitude;
	else if (player->vzx<-constants->terminal_turning_magnitude)
		player->vzx = -constants->terminal_turning_magnitude;

	if (player->angular_acceleration>constants->terminal_acceleration)
		player->angular_acceleration = constants->terminal_acceleration;
	else if (player->angular_acceleration<-constants->terminal_acceleration)
		player->angular_acceleration = -constants->terminal_acceleration;

	player->az = NFix30Mul(cosine_table[departure_angle],player->angular_acceleration);
	player->ax = NFix30Mul(sine_table[departure_angle],player->angular_acceleration);

	switch(action_flags&(_zoom_in|_zoom_out))
	{
	case _zoom_in: player->sz += ZOOM_MAGNITUDE; break;
	case _zoom_out: player->sz -= ZOOM_MAGNITUDE; break;
	}

	if (player->sz<MINIMUM_ZOOM)
		player->sz = MINIMUM_ZOOM;
	else if (player->sz<MAXIMUM_ZOOM)
		player->sz = MAXIMUM_ZOOM;

	/* now move the player */
	move_player(player,action_flags);
}

static void transform_endpoints(world_endpoint *endpoint,
	world_point *transformed,
	SInt32 ox,
	SInt32 oy,
	SInt32 oz,
	int xy,
	int yz,
	int zx,
	SInt32 c)
{
	sfract			cxy = cosine_table[xy], sxy = sine_table[xy];
	sfract			cyz = cosine_table[yz], syz = sine_table[yz];
	sfract			czx = cosine_table[zx], szx = sine_table[zx];

	if (c<=0 || transformed==NULL)
		return;
	do
	{
		SInt32		x = endpoint->x-ox, y = endpoint->y-oy, z = endpoint->z-oz;

		/* transform xy */
		transformed->x = NFix30Mul(x,cxy)+NFix30Mul(y,sxy);
		y = NFix30Mul(y,cxy)-NFix30Mul(x,sxy);
		/* transform yz */
		transformed->y = NFix30Mul(y,cyz)+NFix30Mul(z,syz);
		z = NFix30Mul(z,cyz)-NFix30Mul(y,syz);
		/* transform zx */
		transformed->z = NFix30Mul(z,cyz)+NFix30Mul(transformed->x,syz);
		transformed->x = NFix30Mul(transformed->x,cyz)-NFix30Mul(z,syz);

		++endpoint, ++transformed;
	} while(--c);
}

static world_point	*transformed_buffer = NULL;
static SInt32		transformed_count = 0;

#define SafeMoveTo(h,v)				do{int _tmpx=h,_tmpy=v; if(_tmpx<-16000) \
	_tmpx=-16000; else if(_tmpx>16000) _tmpx=16000; if(_tmpy<-16000) \
	_tmpy=-16000; else if(_tmpy>16000) _tmpy=16000; MoveTo(_tmpx,_tmpy);}while(0)
#define SafeLineTo(h,v)				do{int _tmpx=h,_tmpy=v; if(_tmpx<-16000) \
	_tmpx=-16000; else if(_tmpx>16000) _tmpx=16000; if(_tmpy<-16000) \
	_tmpy=-16000; else if(_tmpy>16000) _tmpy=16000; LineTo(_tmpx,_tmpy);}while(0)

static void render_screen(world_struct *world)
{
	Rect			clipping_rect;
	RGBColor		color;

	GDHandle		old_device;
	CGrafPtr		old_port;

	world_player	*player = world->local_player+world->players;
	world_triangle	*triangle = world->triangles;
	SInt32			x, sx, sy, sd, half_screen_x, half_screen_y;
	sfixed			sz;

	GetGWorld(&old_port,&old_device);
	SetGWorld(world_port,NULL);
	EraseRect(&port_rectangle);

	half_screen_x = RECT_WIDTH(&port_rectangle)+1;
	half_screen_y = RECT_HEIGHT(&port_rectangle)+1;

	if ((x=world->triangle_count)<=0 || world->endpoint_count<=0
			|| transformed_buffer==NULL)
		return;
	if ((sd=player->sd)==0) /* integerized 1.0 cannot be zero */
		sd = player->sd = 1;
	sx = player->sx*2, sy = player->sy*2, sz = NFix16Div(player->sz,sd);

	transform_endpoints(world->endpoints,transformed_buffer,
		player->px,player->py,player->pz,
		player->xy,player->yz,player->zx,
		world->endpoint_count);

	/* the camera is treated as a plate-like reverse-transformed plane at z=0 */
	do
	{
		world_point		p0 = transformed_buffer[triangle->e0];
		world_point		p1 = transformed_buffer[triangle->e1];
		world_point		p2 = transformed_buffer[triangle->e2];

		/* for points that go behind the camera... */
		clipping_rect = port_rectangle;
		if (p0.z<0) /* if ((cz=p0.z-sd)<0) */
		{ /* interpolate the points toward z=0 and make a clipping rectangle */
			if (p1.z<=0 && p2.z<=0)
				continue; /* completely clipped off */
			if (p0.x<p1.x || p0.x<p2.x)
				if (p1.x<p2.x)
					clipping_rect.left = imax(NMulDiv(NMulDiv(p0.x-p1.x,p1.z,p0.z)+p1.x,sx,p0.z)+half_screen_x>>1,port_rectangle.left);
				else
					clipping_rect.left = imax(NMulDiv(NMulDiv(p0.x-p2.x,p2.z,p0.z)+p2.x,sx,p0.z)+half_screen_x>>1,port_rectangle.left);
			else if (p0.x>p1.x || p0.x>p2.x)
				if (p1.x>p2.x)
					clipping_rect.right = imin(NMulDiv(NMulDiv(p0.x-p1.x,p1.z,p0.z)+p1.x,sx,p0.z)+half_screen_x>>1,port_rectangle.right);
				else
					clipping_rect.right = imin(NMulDiv(NMulDiv(p0.x-p2.x,p2.z,p0.z)+p2.x,sx,p0.z)+half_screen_x>>1,port_rectangle.right);

			if (p0.y<p1.y || p0.y<p2.y)
				if (p1.y<p2.y)
					clipping_rect.left = imax(NMulDiv(NMulDiv(p0.y-p1.y,p1.z,p0.z)+p1.y,sx,p0.z)+half_screen_x>>1,port_rectangle.top);
				else
					clipping_rect.left = imax(NMulDiv(NMulDiv(p0.y-p2.y,p2.z,p0.z)+p2.y,sx,p0.z)+half_screen_x>>1,port_rectangle.top);
			else if (p0.y>p1.y || p0.y>p2.y)
				if (p1.y>p2.y)
					clipping_rect.right = imin(NMulDiv(NMulDiv(p0.y-p1.y,p1.z,p0.z)+p1.y,sx,p0.z)+half_screen_x>>1,port_rectangle.bottom);
				else
					clipping_rect.right = imin(NMulDiv(NMulDiv(p0.y-p2.y,p2.z,p0.z)+p2.y,sx,p0.z)+half_screen_x>>1,port_rectangle.bottom);
		}

		ClipRect(&clipping_rect);

		if ((p0.z=NFix16Mul(p0.z,sz))==0)
			p0.z = 1;
		if ((p1.z=NFix16Mul(p1.z,sz))==0)
			p1.z = 1;
		if ((p2.z=NFix16Mul(p2.z,sz))==0)
			p2.z = 1;
		color.red = triangle->basic.cr*257;
		color.green = triangle->basic.cg*257;
		color.blue = triangle->basic.cb*257;
		RGBForeColor(&color);
		SafeMoveTo(NMulDiv(p0.x,sx,p0.z)+half_screen_x>>1,NMulDiv(p0.y,sy,p0.z)+half_screen_y>>1);
		SafeLineTo(NMulDiv(p1.x,sx,p1.z)+half_screen_x>>1,NMulDiv(p1.y,sy,p1.z)+half_screen_y>>1);
		SafeLineTo(NMulDiv(p2.x,sx,p2.z)+half_screen_x>>1,NMulDiv(p2.y,sy,p2.z)+half_screen_y>>1);
		SafeLineTo(NMulDiv(p0.x,sx,p0.z)+half_screen_x>>1,NMulDiv(p0.y,sy,p0.z)+half_screen_y>>1);
	} while(++triangle,--x);

	color.red = color.green = color.blue = 0;
	RGBForeColor(&color);

	MoveTo(2,10);
	TextFont(4);
	TextSize(9);
	DrawText(temporary,0,sprintf(temporary,"p[%.3f %.3f %.3f]", (double)player->px/sd,(double)player->py/sd,(double)player->pz/sd));
	MoveTo(2,20);
	DrawText(temporary,0,sprintf(temporary,"r[%.3f %.3f %.3f]", (double)player->px*(360.0/42967296.0),(double)player->py*(360.0/42967296.0),(double)player->pz*(360.0/42967296.0)));

	SetGWorld(old_port,old_device);
}

static void update_screen(void)
{
	RGBColor		old_fore, old_back, color;
	GrafPtr			old_port;

	if (world_window==NULL)
		return;

	GetPort(&old_port);
	SetPortWindowPort(world_window);

	GetForeColor(&old_fore);
	GetBackColor(&old_back);

	color.red = color.green = color.blue = 0;
	RGBForeColor(&color);
	color.red = color.green = color.blue = 0xFFFF;
	RGBBackColor(&color);

	CopyBits((BitMapPtr)*world_port->portPixMap,
		(BitMapPtr)*GetWindowPort(world_window)->portPixMap,
		&port_rectangle,&dest_rectangle,srcCopy,NULL);

	RGBForeColor(&old_fore);
	RGBBackColor(&old_back);

	SetPort(old_port);
}

void initialize_world(world_struct *world)
{
	world->time_frame = 0;
	world->random_seed = TickCount();
}

static void update_world(world_struct *world, UInt16 advance_time)
{
	UInt32			action_flags[1];

	action_flags[0] = update_keys(local_keys); /* was action_flags[1] until now!; 2008.04.17 10:36 EST */
	if (world->player_count!=1 || world->local_player!=0)
		DebugStr("\pupdate_world: Can only handle one player!");
	do
	{
		update_player(world->players,action_flags[0]);
		++world->time_frame;
	} while(--advance_time);

	render_screen(world);
	update_screen();
}

void run_world(world_struct *world)
{
	UInt32			previous_time = TickCount()-8, current_time = 0;

	if (transformed_buffer!=NULL)
		{free(transformed_buffer); transformed_buffer = NULL;}
	if (world->endpoint_count>0)
		if ((transformed_buffer=malloc(sizeof(*transformed_buffer)*world->endpoint_count))==NULL)
		{
			fputs("### run_world: failed to allocate transformed endpoint buffer",stderr);
			return;
		}
	do
	{
		current_time = TickCount();
		update_world(world,current_time-previous_time);
		previous_time = current_time;
	} while(!Button());

	if (transformed_buffer!=NULL)
		{free(transformed_buffer); transformed_buffer = NULL;}

	if (world_window!=NULL)
		InvalWindowRect(world_window,&dest_rectangle);
}
