/*
 sbcp_compiler.c
  contains pretty much just an assembler.

  Version: sbcp 0.1.7

  Created: November 29, 2007 5:27 AM EDST
  Revised: January 12, 2008 10:12 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
 plans:
	if expression then -> make label _#ifxxx
	else -> make label _#elxxx
	end if -> make label _#fixxx

	...XDB Code?
*/

#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
/*#include <math.h>*/
/* math.h and fp.h conflict with each other! */
#include <fp.h>

#include "sbcp_core.h"
#include "sbcp_utils.h"

#ifdef env68k
# pragma segment sbcp
#endif

/*
	!directive
	label: #numeric ;comment
	@macro (operands)
*/

/* ---------- constants */

#ifndef MINUS_CHAR /* MacRoman: '\320' */
# define MINUS_CHAR						'-'
#endif
#ifndef DBLSP_CHAR /* MacRoman: '\312' (double space) */
# define DBLSP_CHAR						' '
#endif

/* ---------- macros */

#define make_cerror(x)					((char*)(1-(int)x))

/* 0-9_A-Za-z (doesn't check to see if first char is a digit) */
#define scan_to_identifier_end(s,e)		do{_uint8 _c=*s; if((_c+=-'0')>(_uint8)('9'-'0') \
	&&(_c+='0'-'_')!=0&&(_c+='_'-'A')>('Z'-'A')&&(_c+='A'-'a')>('z'-'a')) break;}while(++s<e)
/* A-Za-z */
#define scan_to_word_end(s,e)			do{_uint8 _c=*s; if((_c+=-'A')>('Z'-'A') \
	&&(_c+='A'-'a')>('z'-'a')) break;}while(++s<e)

#if DBLSP_CHAR!=' '
# define skip_whitespace(s,e)			do{if(s>=e) break; c=*s++;}while((unsigned)c<=' '||c==DBLSP_CHAR)
#else
# define skip_whitespace(s,e)			do{if(s>=e) break; c=*s++;}while((unsigned)c<=' ')
#endif

/* ---------- code */

static _uint8 *lower_pstrcpy(_uint8 *_d, const void *_s, int z)
{
	const _uint8	*s = _s;
	_uint8			*d;

	if ((d=_d)==NULL)
		return NULL;
	if (z<=0)
		*d++ = 0;
	else
		{*d++ = z; do{*d++ = tolower(*s++);} while(--z!=0);}

	return _d;
}

static int match_string(const char *a, const char *s, int z)
{
	char			c;

	if (z==0)
		return 0;
	do
	{
		if ((c=*a++-tolower(*s++))!=0)
			return c;
	} while(--z!=0);

	return 0;
}

static int strlist_search(const char *s,
	int z,
	const _uint8 *l,
	int n,
	boolean r(const _uint8*,int)) /* an optional routine for further verification */
{
	_uint8			t, f;
	int				i = -1;

	if (z==0 || n<=0)
		return -1;
	f = tolower(*s++);
	if (z<0)
	{ /* this side (when the length parameter is negative) sees if the length is sufficient */
		z = -z;
		do
		{
			++i;
			if ((t=*l++)>=z && *l==f) /* length and first char check */
				if (match_string(1+(const char*)l,s,z-1)==0 && (r==NULL || r(-1+l,z)))
					return i;
			l += t;
		} while(--n!=0);
	} else /* otherwise the length has to match */
		do
		{
			++i;
			if ((t=*l++)==z && *l==f) /* length and first char check */
				if (match_string(1+(const char*)l,s,z-1)==0 && (r==NULL || r(-1+l,z)))
					return i;
			l += t;
		} while(--n!=0);

	return -1;
}

/* 16 bit lengths */
static int wstrlist_search(const char *s,
	int z,
	const _uint8 *l,
	int n,
	boolean r(const _uint8*,int)) /* an optional routine for further verification */
{
	_uint8			t, f;
	int				i = -1;

	if (z==0 || n<=0)
		return -1;
	f = tolower(*s++);
	if (z<0)
	{ /* this side (when the length parameter is negative) sees if the length is sufficient */
		z = -z;
		do
		{
			++i;
			if ((t=*((UInt16*)l)++)>=z && *l==f) /* length and first char check */
				if (match_string(1+(const char*)l,s,z-1)==0 && (r==NULL || r(-1+l,z)))
					return i;
			l += t;
		} while(--n!=0);
	} else /* otherwise the length has to match */
		do
		{
			++i;
			if ((t=*((UInt16*)l)++)==z && *l==f) /* length and first char check */
				if (match_string(1+(const char*)l,s,z-1)==0 && (r==NULL || r(-1+l,z)))
					return i;
			l += t;
		} while(--n!=0);

	return -1;
}

/* ---------- numeric processing */

enum
{
	_cci_version,				/* value of the current version number (in BCD) */
	_cci_date,					/* value of the current date/time in seconds */
	_cci_pi,					/* constant pi = atan(1)*4 */
	_cci_gr,					/* constant golden ratio = 1/(golden ratio-1) */
	_cci_c,						/* constant c = speed of light in a vacuum m/s */
	_cci_e,						/* constant e = natural logarithm base */
	_cci_sqrt2,					/* square root of two */
	_cci_12root2,				/* twelfth-root of two */
	_cci_log2ten,				/* log-base-two of ten */

	_compiler_const_id_count
};

const _uint8	_compiler_const_ids[] =
{ /* \p doesn't work in this context */
	"\007version"
	"\004date"
	"\002pi"
	"\002gr"
	"\001c"
	"\001e"
	"\005sqrt2"
	"\00712root2"
	"\007log2ten"
};

#define EMPTY_VALUE						0

static const char *do_numeric(const char *s, const char *e, double *_v)
{
	double			v;
	const char		*p;
	_uint8			c;

	if (_v!=NULL)
		*_v = EMPTY_VALUE;
	if (s>=e)
		return s;
/*
	c = *s++;
	// allow for #number (but disallow special opcodes)
	if ((c+=-'0')<=(_uint8)('9'-'0') || (((c+='0'-'-')==0 || (c+='-'-'+')==0
			|| (c+='+'-'.')==0) && (s<e || (_uint8)(*s-'0')<=(_uint8)('9'-'0'))))
		return sbcp_parse_number(--s,e,_v,true);
	p = --s;
*/
	p = s;
	scan_to_identifier_end(s,e);

	switch(strlist_search(p,s-p,_compiler_const_ids,_compiler_const_id_count,NULL))
	{
	case _cci_version: v = SBCP_VERSION; break;
	case _cci_date: v = time(NULL); break;

	case _cci_pi: v = 3.141592653589793238462643383279503; break;
	case _cci_gr: v = 1.618033988749894848204586834365638; break;
	case _cci_c: v = 299792458.0; break;
	case _cci_e: v = 2.718281828459045235; break;
	case _cci_sqrt2: v = 1.414213562373095048801688724209698; break;
	case _cci_12root2: v = 1.059463094359295264561825294946342; break;
	case _cci_log2ten: v = 3.321928094887362347870319429489390; break;

	default:
		return make_cerror(_sbcc_undefined);
	}

	if (_v!=NULL)
		*_v = v;

	return s;
}

/* ---------- macro processing */

static const char *do_macro(const char *s, const char *e, _uint8 **d)
{ /* not implemented, so far */
#pragma unused(s,e,d)
	return NULL;
}

/* ---------- directive processing */

static const char *do_directive(const char *s, const char *e, _uint8 **d)
{ /* not implemented, so far */
#pragma unused(s,e,d)
	return NULL;
}

/* ---------- processing of other identifiers */

static const char *identifier_to_index(const char *s, const char *e, int *v)
{
	if (s>=e)
		return NULL;

	if (v!=NULL)
		*v = INT_MIN; /* invalid value */
	return NULL;
}

static _uint8	*mnemonic_list = NULL;
static long		mnemonics_length = 0, mnemonics_count = 0;

/* label generation and mnemonics */
static const char *do_mnemonic(const char *s,
	const char *e,
	_uint8 **d,
	_uint32 o)
{
#pragma unused(d)
	_uint8			*t;
	const char		*p = s;
	int				z;

	scan_to_identifier_end(s,e);
	if (s-p>255)
		return make_cerror(_sbcc_name_too_long);
	if (s<e && *s==':')
	{ /* label */
		if (wstrlist_search(p,z=++s-p,mnemonic_list,mnemonics_count,NULL)>=0)
			return make_cerror(_sbcc_duplicate_label);
		/* add the label */
		/* this will actually add two ids: first, the name; then the code offset (in binary) */
		if ((t=realloc(mnemonic_list,mnemonics_length+z+4+4))==NULL) /* lengths plus length _uint16's */
			return make_cerror(_sbcc_memory_error);
		mnemonic_list = t, mnemonics_count += 2;
		t += mnemonics_length;

		/* these are Pascal strings in lowercase */
		*t++ = 0; lower_pstrcpy(t,p,z); t += z+=2;
		*t++ = 0; *t++ = 4; _encodexint32(t,o);
		return s;
	}

	return NULL;
}

#define INSTRUCTION_COUNT				(sizeof(_sbcp_opcodes)/sizeof(_sbcp_opcodes[0]))

const _uint8		*retin_offset;

static boolean get_colon_offset(const _uint8 *l, int z)
{
	retin_offset = l, ++z;
	return *l==z && l[z]==':';
}

/* refer to sbcp_debug.c and sbcp_core.h for instructions and operands */
static const char *do_instruction(char c,
	const char *s,
	const char *e,
	_uint8 **_d,
	_uint32 o)
{
	_uint8			*d = *_d;
	const char		*l = *_sbcp_opcodes, *p = s;
	double			v;
	int				x = INSTRUCTION_COUNT, z; /* instruction opcode count */

	scan_to_word_end(s,e);
	if ((unsigned)((z=s-p)-1)>6) /* length outside of [1,7] */
		return NULL;
	do
	{
		if (*l==c && l[z]=='\0' && match_string(p,1+l,z))
		{
			x = -x, x += INSTRUCTION_COUNT;
			switch(l[z+1])
			{
			/*case _sbci_none: *d++ = x, *d++ = *d++ = *d++ = 0; break;*/
			case _sbci_value:
				if ((s=sbcp_parse_number(s,e,&v,false))==NULL)
					return NULL;
				if (fabs(v+.5)>8388607.5)
					return make_cerror(_sbcc_value_error);
				*d++ = x; _encodexint24(d,v);
				break;
			case _sbci_frame_index:
			case _sbci_frame_optional:
				/* this may require an implementation for stack frame variables... */
				break;
			case _sbci_offset:
				skip_whitespace(s,e);
				if (*(_uint8*)(p=--s)=='*') /* near offset */
				{
					if ((s=sbcp_parse_number(++s,e,&v,false))==NULL)
						return NULL;
					if (fabs(v+.5)>8388607.5)
						return make_cerror(_sbcc_value_error);
					*d++ = x; _encodexint24(d,v);
					break;
				}
				scan_to_identifier_end(s,e);
				retin_offset = NULL;
				if (wstrlist_search(p,p-s,mnemonic_list,mnemonics_count,get_colon_offset)>=0
						&& retin_offset!=NULL)/* || *retin_offset!=1+s-p) */
				{
					const _uint8	*t = retin_offset;
					_uint32			b;

					t += *((_uint16*)t)++;
					if (*((_uint16*)t)++!=4)
						return make_cerror(_sbcc_internal_error); /* should be four bytes! */

					_decodeuint32(t,b);
					*d++ = x; _encodexint24(d,b-o);
				} else
					return make_cerror(_sbcc_undefined);
				break;
			case _sbci_float24:
				if ((s=sbcp_parse_number(s,e,&v,true))==NULL)
					return NULL;
				if (!sbcp_float24_range_check(v))
					return make_cerror(_sbcc_value_error);
				*d++ = x; _encodexint24(d,sbcp_make_float24(v));
				break;
			case _sbci_register:
				break;
			case _sbci_displaced:
				break;
			case _sbci_displaced_index:
				break;
			case _sbci_field:
				break;
			case _sbci_field_index:
				break;
			case _sbci_cache_table:
				break;
			}
			*_d = d;
			return s;
		}
		l += sizeof(*_sbcp_opcodes);
	} while(--x);

	return NULL;
}

/* ---------- symbol processing */

/*
   Instruction(s)	   Symbol		   Operand(s) or Comment
	exg:				><						
	popr:				=
	pshd (current):		{
	popd (current):		}
	orc:				=|				number
	andc:				=&				number
	addc:				=+				number
	neg, addc:			=-				number
	cmpd:				=~				number
	or:					|
	or, snz:			||
	and:				&
	and, snz:			&&
	exor:				^
	exor:				^^
	snz:				~
	sge:				>=
	sle:				<=
	slt:				<
	sgt:				>
	neg:				-			; (hyphen) - minus character also on some systems
	add:				+
	mul:				*
	div:				/
	abs:				++
	scal:				<<
	neg, scal:			>>			; for literals, (b<0; a<<b) is more efficient than a<<-(b)
*/
static const char *do_symbol(char c, const char *s, const char *e, _uint8 **_d)
{
	_uint8			*d = *_d;
	int				x;

	if (s>=e || (unsigned)*s<=' ' || *s==';')
	{
		if ((c+=-'~')==0)
			*d++ = _ins_snz, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='~'-'<')==0)
			*d++ = _ins_slt, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='<'-'>')==0)
			*d++ = _ins_sgt, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='>'-'+')==0)
			*d++ = _ins_add, *d++ = 0x80, *d++ = *d++ = 0;
#if MINUS_CHAR!='-'
		else if ((c+='+'-MINUS_CHAR)==0 || (c+=MINUS_CHAR-'-')==0)
#else
		else if ((c+='+'-'-')==0)
#endif
			*d++ = _ins_neg, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='-'-'*')==0)
			*d++ = _ins_mul, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='*'-'/')==0)
			*d++ = _ins_div, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='/'-'|')==0)
			*d++ = _ins_or, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='|'-'&')==0)
			*d++ = _ins_and, *d++ = 0x80, *d++ = *d++ = 0;
		else if ((c+='&'-'^')==0)
			*d++ = _ins_exor, *d++ = 0x80, *d++ = *d++ = 0;
		else
			return NULL;
		goto end;
	}
#if MINUS_CHAR!='-'
	if ((c+=-'-')==0 || (_uint8)c==(_uint8)(MINUS_CHAR-'-'))
#else
	if ((c+=-'-')==0)
#endif
	{
		if ((unsigned)(*s-'0')>(_uint8)9)
			{*d++ = _ins_neg, *d++ = 0x80, *d++ = *d++ = 0; goto end;}
	} else if ((c+='-'-'+')==0)
		if (*s=='+')
			{*d++ = _ins_abs, *d++ = 0x80, *d++ = *d++ = 0, ++s; goto end;}
		else if ((unsigned)(*s-'0')>(_uint8)9)
			{*d++ = _ins_add, *d++ = 0x80, *d++ = *d++ = 0; goto end;}

	/* "<>" removed to eliminate confusion with "><" */
	if ((c+='+'-'<')==0)
		if (*s=='<')
			{*d++ = _ins_scal, *d++ = 0x80, *d++ = *d++ = 0, ++s; goto end;}
		else if (*s=='=')
			{*d++ = _ins_sle, *d++ = 0x80, *d++ = *d++ = 0, ++s; goto end;}
		else
			{*d++ = _ins_slt, *d++ = 0x80, *d++ = *d++ = 0; goto end;}
	if ((c+='<'-'>')==0)
		if (*s=='>')
			{*d++ = _ins_neg, *d++ = _ins_scal, *d++ = 0x80, *d++ = *d++ = 0, ++s; goto end;}
		else if (*s=='<')
			{*d++ = _ins_exg, *d++ = 0x80, *d++ = *d++ = 0, ++s; goto end;}
		else if (*s=='=')
			{*d++ = _ins_sge, *d++ = 0x80, *d++ = *d++ = 0, ++s; goto end;}
		else
			{*d++ = _ins_sgt, *d++ = 0x80, *d++ = *d++ = 0; goto end;}
	if ((c+='>'-'=')==0)
		if (*s=='|')
		{ /* orc requires an operand */
			if ((s=identifier_to_index(++s,e,&x))==NULL)
				return make_cerror(_sbcc_no_operand);
			*d++ = _ins_orc; _encodexint24(d,x); goto end;
		} else if (*s=='&')
		{ /* andc requires an operand */
			if ((s=identifier_to_index(++s,e,&x))==NULL)
				return make_cerror(_sbcc_no_operand);
			*d++ = _ins_andc; _encodexint24(d,x); goto end;
		} else if (*s=='+')
		{ /* addc requires an operand */
			if ((s=identifier_to_index(++s,e,&x))==NULL)
				return make_cerror(_sbcc_no_operand);
			*d++ = _ins_addc; _encodexint24(d,x); goto end;
		} else if (*s=='-')
		{ /* subc mnemonic requires an operand */
			if ((s=identifier_to_index(++s,e,&x))==NULL)
				return make_cerror(_sbcc_no_operand);
			/**d++ = _ins_pshz; *d++ = _ins_exg; *d++ = _ins_sub;*/
			*d++ = _ins_neg, *d++ = 0x80, *d++ = *d++ = 0;
			*d++ = _ins_addc; _encodexint24(d,x); goto end;
		} else if (*s=='~')
		{ /* cmpc requires an operand */
			if ((s=identifier_to_index(++s,e,&x))==NULL)
				return make_cerror(_sbcc_no_operand);
			*d++ = _ins_cmpd; _encodexint24(d,x); goto end;
		} else
			{*d++ = _ins_popr, *d++ = 0x80, *d++ = *d++ = 0; goto end;}

	if ((c+='='-'{')==0)
		{*d++ = _ins_pshd; _encodexint24(d,0x800000L);}
	else if ((c+='{'-'}')==0)
		{*d++ = _ins_popd; _encodexint24(d,0x800000L);}
	else if ((c+='}'-'*')==0)
		*d++ = _ins_mul, *d++ = 0x80, *d++ = *d++ = 0;
	else if ((c+='*'-'/')==0)
		*d++ = _ins_div, *d++ = 0x80, *d++ = *d++ = 0;
	else if ((c+='/'-'|')==0)
		*d++ = _ins_or, *d++ = 0x80, *d++ = *d++ = 0;
	else if ((c+='|'-'&')==0)
		*d++ = _ins_and, *d++ = 0x80, *d++ = *d++ = 0;
	else if ((c+='&'-'^')==0)
		*d++ = _ins_exor, *d++ = 0x80, *d++ = *d++ = 0;
	else
		return NULL;
end:
	*_d = d;
	return s;
}

/* ---------- main compiler */

/* this should be larger than the maximum size for each code addition */
#define MINIMUM_CODE_ALLOCATION			256

/* this may require allocations, both of stack space and realloc space; the
  alternative of writing to the stack would of course require copying the result
  on every re-buffering (which would probably be slower) */
sbcc_error sbcp_compile(const char *text,
	long *out_length,
	_uint8 **out_data,
	long *err_offset)
{
	_uint8			buf_sub[MINIMUM_CODE_ALLOCATION];

	_uint8			*data = NULL, *d = NULL, c;
	const char		*s, *e, *t;
	double			v, n;
	_uint32			x;
	long			length, allocated = 0;
	sbcc_error		err = _sbcc_no_error;

	if ((s=text)==NULL || out_length==NULL)
		return _sbcc_no_error;
	if ((length=*out_length)<=0)
		{*out_length = 0; return _sbcc_no_error;}
	e = length+s;
	do
	{
		t = s;
		if (out_data==NULL)
			d = buf_sub; /* so the code gets written to a safe (but unused) location */
		else if (allocated-(d-data)<MINIMUM_CODE_ALLOCATION)
		{
			long		offset = d-data;

			if ((t=realloc(data,allocated+=MINIMUM_CODE_ALLOCATION*2))==NULL)
				{err = _sbcc_memory_error; break;}
			d = offset+(data=(_uint8*)t);
		}
		/*if ((c=*s++)==0 || (c+=-'\t')==0 || (c+='\t'-'\r')==0 || (c+='\r'-'\f')==0 || (c+='\f'-'\n')==0)
			continue;*/
		if ((unsigned)(c=*s++)<=' ') /* faster (but somewhat dumber) approach to skip whitespace */
			continue;
#if DBLSP_CHAR!=' '
		if (c==(_uint8)DBLSP_CHAR)
			continue;
#endif
		if ((c+=-';')==0)
		{ /* comment; skip line */
			do
			{
				if (s>=e)
					goto end;
				c = *s++;
			} while((c+=-'\r')!=0 && (c+='\r'-'\f')!=0 && (c+='\f'-'\n')!=0);
		} else if ((c+=';'-'#')==0)
		{
			if ((t=do_numeric(s,e,&v))!=NULL)
				goto encode_number;
		} else if ((c+='#'-'@')==0)
			t = do_macro(s,e,&d);
		else if ((c+='@'-'!')==0)
			t = do_directive(s,e,&d);
		else if ((c+='!'-'.')==0 || (c+='.'-'-')==0 || (c+='-'-'+')==0
#if MINUS_CHAR!='-'
				|| c==(_uint8)(MINUS_CHAR-'+')
#endif
				|| (unsigned)(c+='+'-'0')<=(_uint8)9)
		{
			if ((t=sbcp_parse_number(-1+s,e,&v,true))==NULL)
				goto try_symbol;
		encode_number:
			float24_to_double(x=sbcp_make_float24(v),&n);
			if (n==v)
				{*d++ = _ins_pshi; _encodexint24(d,x);}
			else
			{ /* encode it in near code */
				*d++ = _ins_pshn, *d++ = *d++ = 0, *d++ = 8;
				*d++ = _ins_jn, *d++ = *d++ = 0, *d++ = 8;
				_encodedouble(d,v);
			}
		} else if ((c+='0'-'_')==0 || (unsigned)(c+='_'-'A')<=(_uint8)('Z'-'A')
				|| (unsigned)(c+='A'-'a')<=(_uint8)('z'-'a'))
		{
			if ((t=do_mnemonic(-1+s,e,&d,d-data))==NULL)
				t = do_instruction(c,s,e,&d,d-data);
		} else
		{
		try_symbol:
			if ((t=do_symbol(s[-1],s,e,&d))==NULL)
			{ /* could be a recognizable char in a bad place */
				if ((c+='a'-':')!=0)
					{err = _sbcc_bad_char; break;}
			}
		}
		/* if all else failed; assume it's the user's fault (at least somebody...) */
		if ((unsigned long int)(t-1)>=~(unsigned long int)_sbcc_name_too_long)
			{err = (sbcc_error)(1-(unsigned long int)t); break;} /* NULL comes out _syntax_error*/
		if (t<s) /* allow an endless loop to the mystical infinite eternity? */
			{err = _sbcc_internal_error; break;}
		s = t;
	} while(s<e);
end:
	/* deallocate all storage that is valid only for this context */
	if (mnemonic_list!=NULL)
		{free(mnemonic_list); mnemonic_list = NULL, mnemonics_length = mnemonics_count = 0;}

	*out_length = 0;
	if (out_data!=NULL)
		if (d-data==0)
			{free(data); *out_data = NULL;}
		else if ((*out_data=realloc(data,*out_length=d-data))==NULL) /* minimize allocation */
			*out_data = data;

	if (err_offset!=NULL)
		if ((*err_offset=s-text-1)<0)
			*err_offset = 0;
	return err;
}
