/*
 sbcp_core.h
  stack based core prototype core include

 Version: 0.1.7
 Created: November 29, 2007 12:02 PM EDST
 Revised: January 14, 2008 5:38 AM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Sunday, December 2, 2007 3:36 AM
A.P.: _ins_sub is now _ins_neg (and comes before _ins_add).
Friday, December 21, 2007 11:02 AM
  _ins_neg-_ins_scal moved, frame index for _ins_add, _ins_mul, _ins_div and _ins_scal.
Sunday, January 13, 2007 10:19 PM
  more revisions to the instruction set...
*/

#ifndef __SBCP_CORE_H
# define __SBCP_CORE_H

# ifndef __CONDITIONALMACROS__
#  include <ConditionalMacros.h>
# endif

# include "sbcp.h"

# if PRAGMA_ONCE
#  pragma once
# endif

# ifdef __cplusplus
   extern "C" {
# endif

# if PRAGMA_IMPORT
#  pragma import on
# endif

/* ---------- constants */

enum
{ /* exception vector indices */
	_ev_access_fault,		/* address cannot be accessed for read/write */
	_ev_write_fault,		/* write to read-only space */
	_ev_illegal_opcode,		/* opcode is bad or unimplemented */
	_ev_illegal_address,	/* odd address (or NULL address in exception handler) */

	_ev_stack_overflow,		/* DS and/or LS went below floor */
	_ev_stack_underflow,	/* DS and/or LS went above ceiling */
	_ev_outside_frame,		/* stack frame index/offset is out of bounds */
	_ev_non_integeral,		/* value has a fraction; operation calls for an integer */

	_ev_execution_fault,	/* last resort (worst case scenario) */
	_ev_privileged,			/* operation is priviledged (can't just muck around) */
	_ev_math_exception,		/* control and status bit (division by zero, etc.) */
	_ev_break_vector		/* debugging vector - break bit (instruction and SR) */

	_ev_trap_vector			/* debugging vector - trap instruction */
};

enum
{ /* FPR bits ([control.16] [status.8] [condition.8]) */
  /* condition bits are set to zero prior to math operations (status bits may subsume) */
	_fprn_zero				= 0x00000001L, /* result is zero */
	_fprn_minus				= 0x00000002L, /* result is negative */
	_fprn_infinite			= 0x00000004L, /* result is +-INF */
	_fprn_nan				= 0x00000008L, /* result is +-NAN */
	_fprn_underflow			= 0x00000010L, /* result is zero, exponent overflow */
	_fprn_overflow			= 0x00000020L, /* result is infinite, exponent overflow */
	_fprs_dbz				= 0x00000040L, /* divide by zero */
	_fprs_inexact			= 0x00000080L, /* precision was lost */

  /* status bits; with matching control bit, throws exception on next math operation */
	_fprs_inexact_a_error	= 0x00000100L, /* only thrown by functions that need precision */
	_fprs_inexact_b_error	= 0x00000200L, /* only thrown by functions that need precision */
	_fprs_operand_error_a	= 0x00000400L, /* non-trig: sqrt(-n), ln(-n), etc. */
	_fprs_operand_error_b	= 0x00000800L, /* trig: sin(+-INF), asin(2), etc. */
	_fprs_underflow			= 0x00001000L, /* zero due to exponent overflow */
	_fprs_overflow			= 0x00002000L, /* infinite due to exponent overflow */
	_fprs_divide_by_zero	= 0x00004000L, /* (n/zero) */
/*	_fprs_signaling_nan		= 0x00008000L, */

  /* control bits control whether an exception is actually thrown _after_ a math error */
	_fprc_inexact_a_error	= 0x00010000L,
	_fprc_inexact_b_error	= 0x00020000L,
	_fprc_operand_error_a	= 0x00040000L,
	_fprc_operand_error_b	= 0x00080000L,
	_fprc_underflow			= 0x00100000L,
	_fprc_overflow			= 0x00200000L,
	_fprc_divide_by_zero	= 0x00400000L,
/*	_fprc_signaling_nan		= 0x00800000L,*/

	_fprc_rounding_prec		= 0x03000000L,
	_fprc_rounding_prshift	= 24,

	_fprn_mask				= 0x000000FFL,
	_fprs_mask				= 0x0000FF00L,
	_fprc_mask				= 0x03FF0000L,
};

enum
{ /* opcodes */
/*
  Special Operands:
	Float24             [sign.1] [exponent.7]-63, 1. [mantissa.16]
                        NOTE: zero is zero but NAN/INF is not represented.

	Field8115           [sign.1] [instance.7] [offset.s8] [update.1]
                           [add_dr.1] [add_zr.1] [width.5]
                        NOTE: `width' of zero is interpreted as 32.
                           `offset' is in bits
                           `update' adds ZR to DR after completion

	Displaced16         [update.1] [instance.7], [offset.s16]
                           `update' displaces the amount beforehand


  Condition codes are shown by their starting letter.
	The that letters may appear in parenthesis indicate whether they are always
	turned on (capital), or off (lowercase).

  Instances are used for direct reads/writes.  There are 128 instances.

  The stack frame top is stored in the link frame.

  "no-ops" with a negative value may represent encoded labels.  Positives
	represent line incrementation (in code).

  For stack frames, index SINT24MIN represents the current, lowest entry;
	positive indices represent parameters to the caller, and negatives point
	within the frame, itself.

  The instruction set is uniform, 32 bits in size.

  The first byte for instructions consists of:
	[break.1]           The instruction is set to break.
	[opnumber.7]        One of the values described in the enum below.

  Any number of mnemonic operations can be achieved:
	remove one (test)   popd
	repeat one (test)   pshd
	boolean OR          or, snz
	boolean AND         and, snz
	bit NOT             pshi -1, exor
	subtract            neg, add
	bitmask insert      pshd xxx, exor, and, exor xxx
*/

   /* 0bx0000000	Name/Description				 Operands					*/
	_ins_nop,	/* no-operation						[ linecount.s24 ]			*/
				/*	primarily used for source info; linecounts<0 seen as zero	*/

	_ins_ads,	/* adjust data stack				[ displacement.s24 ]		*/
				/*	adds displacement<<3 to the data stack position				*/

	_ins_pshn,	/* push near						[ offset.s24 ]				*/
				/*	pushes near double at code offset onto DS; FPR conditions	*/

	_ins_pshi,	/* push immediate					Float24						*/
				/*	as a double, pushed onto data stack; FPR conditions (ni)	*/

	_ins_nea,	/* near effective address			[ offset.s24 ]				*/
				/*	sets RS to a calculated absolute address to the code offset	*/

	_ins_exg,	/* stack entry exchange				[ frameindex.s24 ]			*/
				/*	DS[0] and F(I) are exchanged								*/

	_ins_stop,	/* stop								[ data.24 ]					*/
				/*	full instruction stop										*/

	_ins_ret,	/* return							[ displacement.s24 ]		*/
				/*	pulls link frame, displaces data stack by displacement<<3	*/
   /* 0bx0001000 */
	_ins_pshr,	/* push result						[ frameindex.s24 ]			*/
				/*	stack frame index is set to the value in RS					*/

	_ins_popr,	/* pop result						[ frameindex.s24 ]			*/
				/*	sets RS to value of stack frame index						*/

	_ins_pshd,	/* push stack entry					[ frameindex.s24 ]			*/
				/*	stack entry at top of stack frame -> data stack				*/

	_ins_popd,	/* pop to stack entry				[ frameindex.s24 ]			*/
				/*	pulls data -> stack entry at top of stack frame				*/

	_ins_orc,	/* bitwise OR and condition			[ frameindex.s24 ]			*/
				/*	F(I) |= RS; FPR conditions (im)								*/

	_ins_andc,	/* bitwise AND and condition		[ frameindex.s24 ]			*/
				/*	F(I) &= RS; FPR conditions (im)								*/

	_ins_addc,	/* add and condition				[ frameindex.s24 ]			*/
				/*	F(I) += RS; all FPR conditions apply						*/

	_ins_cmpd,	/* compare and direction			[ frameindex.s24 ]			*/
				/*	RS = relation(F(I),RS); all FPR conditions apply			*/
   /* 0bx0010000 */
	_ins_rdd,	/* read data-double					Displaced16					*/
				/*	reads a double from an instance -> RS						*/

	_ins_rds,	/* read data-single					Displaced16					*/
				/*	reads a 32 bit "single" from an instance -> RS				*/

	_ins_rdb,	/* read data-byte					Displaced16					*/
				/*	reads an 8 bit byte (zero extension) from an instance -> RS	*/

	_ins_rdf,	/* read data-field					Field8115					*/
				/*	reads field as integer -> RS								*/

	_ins_wdd,	/* write data-double				Displaced16					*/
				/*	writes RS as a 64 bit "double" to an instance				*/

	_ins_wds,	/* write data-single				Displaced16					*/
				/*	writes RS as a 32 bit "single" to an instance				*/

	_ins_wdb,	/* write data-byte					Displaced16					*/
				/*	writes RS as an 8 bit byte (unsigned) in instance			*/

	_ins_wdf,	/* write data-field					Field8115					*/
				/*	using RS, writes field as integer							*/
   /* 0bx0011000 */
	_ins_orr	/* bit-or result					[ frameindex.s24 ]			*/
				/*	RS |= F(I); all FPR conditions apply						*/

	_ins_andr	/* bit-and result					[ frameindex.s24 ]			*/
				/*	RS &= F(I); all FPR conditions apply						*/

	_ins_exorr	/* bit-exclusive-or result			[ frameindex.s24 ]			*/
				/*	RS ^= F(I); all FPR conditions apply						*/

	_ins_setri	/* set result to immediate			Float24						*/
				/*	as a double -> RS; FPR conditions (ni)						*/
   /* 0bx0100000 -- note: 'or', 'and', and 'exor' clip to top 52 integer bits	*/
	_ins_or,	/* bitwise or						[ frameindex.s24 ]			*/
				/*	DS[0] |= DS[0]; ++DS; all FPR conditions apply				*/

	_ins_and,	/* bitwise and						[ frameindex.s24 ]			*/
				/*	DS[0] &= DS[0]; ++DS; all FPR conditions apply				*/

	_ins_exor,	/* bitwise exclusive-or				[ frameindex.s24 ]			*/
				/*	F(I) ^= DS[0]; all FPR conditions apply						*/

	_ins_cc,	/* change condition					[am.12], [xm.12]			*/
				/*	FPR ^= (FPR^xm)&am											*/

	_ins_sze,	/* set zero condition equal			[am.12], [mm.12]			*/
				/*	Z = (FPR&am)==mm											*/

	_ins_szn,	/* set zero condition not equal		[am.12], [mm.12]			*/
				/*	Z = (FPR&am)!=mm											*/

	_ins_abs,	/* absolute value					[ frameindex.s24 ]			*/
				/*	set sign bit of F(I) to zero; all SR conditions apply		*/

	_ins_nabs,	/* negative absolute value			[ frameindex.s24 ]			*/
				/*	set sign bit of F(I) to one; all SR conditions apply		*/
   /* 0bx0101000 */
	_ins_add,	/* add								[ frameindex.s24 ]			*/
				/*	DS[0+] += F(I); all SR conditions apply						*/

	_ins_sub,	/* subtract							[ frameindex.s24 ]			*/
				/*	DS[0+] -= F(I); all SR conditions apply						*/

	_ins_mul,	/* multiply							[ frameindex.s24 ]			*/
				/*	DS[0+] *= F(I); all SR conditions apply						*/

	_ins_div,	/* divide							[ frameindex.s24 ]			*/
				/*	DS[0+] /= F(I); all SR conditions apply						*/

	_ins_neg,	/* unary negate						[ frameindex.s24 ]			*/
				/*	invert sign bit of F(I); all SR conditions apply			*/

	_ins_scal,	/* scale by power of two			[ frameindex.s24 ]			*/
				/*	F(I) = scalb(F(I),DS[0]); all SR conditions apply			*/

	_ins_rndz,	/* round toward zero				[ frameindex.s24 ]			*/
				/*	F(I) = trunc(F(I)); all SR conditions apply					*/

	_ins_sqrt,	/* square root						[ frameindex.s24 ]			*/
				/*	F(I) = sqrt(F(I)); all SR conditions apply					*/
   /* 0bx0110000 */
	_ins_jn,	/* jump near						[ offset ]					*/
				/*	jump to offset (always)										*/

	_ins_jnz,	/* jump near on condition zero		[ offset ]					*/
				/*	jump to offset only if zero condition						*/

	_ins_cn,	/* call near						[ offset ]					*/
				/*	adds link, jumps to calculated address; clears the "result"	*/

	_ins_cnz,	/* call near on condition zero		[ offset ]					*/
				/*	calls offset only if zero condition							*/

	_ins_iad,	/* instance addr. dereference		[instance.s8], [offset.s16]	*/
				/*	sets RS to the address of an instance						*/

	_ins_iar,	/* instance addr. reference			[instance.s8], [offset.s16]	*/
				/*	sets an instance to the pulled absolute address				*/

	_ins_iaa,	/* instance address add				[instance.s8], [offset.s16]	*/
				/*	adds byte offset and pulled value to an instance			*/

	_ins_ci,	/* call instance					[instance.s8], [offset.s16]	*/
				/*	calls instance address with offset							*/
	/* 0bx0111000 - Warning: the implementations for these may change! */
	_ins_rct,	/* read cache table (does nothing in the emulator)				*/
				/* reads a 16 byte line for instance zero directly from cache	*/

	_ins_wct,	/* write cache table (does nothing in the emulator)				*/
				/* writes a 16 byte line for instance zero directly to cache 	*/

	_ins_setrg,	/* set register						[index.s8] [reserved.16]	*/
				/* sets an internal register to value of RS						*/

	_ins_getrg,	/* get register						[index.s8] [reserved.16]	*/
				/* sets RS to the value of an internal register					*/

	_ins_setds,	/* set DS							[ address.24 ]				*/
				/* sets DS register to absolute address (useful for exceptions)	*/

	_ins_setpc,	/* set PC							[ address.24 ]				*/
				/* sets PC to absolute address (useful for exceptions)			*/

	_ins_trap,	/* trap								[ data.24 ]					*/
				/* throws a trap vector exception								*/

    _ins_illegal = 127 /* this opcode'll never be defined; operands: undefined */
};

/* ---------- structures */

typedef const char		sbcp_opcode_info_data[8], sbcp_register_name[8];

/* ---------- globals */

/*
  sbcp_opcode_info_data:
    each name <7 chars; the name covers a C-string; blanks are "undefined"

    opcodes that have operands are four bytes in size, otherwise one byte

    each operand uses a char (after the first terminating null) as a type:
      0 = 24 bit signed stack frame index (current represents current value)
      1 = 24 bit value (data)
      2 = 24 bit signed stack frame index (current represents next on stack)
      3 = 24 bit signed offset
      4 = 24 bit Float24
      5 = register record: [which.8] [reserved.16]
      6 = displaced: [sign.1] [instance.7] [offset.s16]
      7 = displaced index: [sign.1] [instance.7] [offset.s8] [reserved.3] [scale.5]
      8 = field: [sign.1] [instance.7] [offset.s16]
      9 = field index: [sign.1] [instance.7] [offset.s8] [reserved.3], [width.5]
     10 = cache table record: [type.8] [offset.16]
*/

enum
{
	_sbci_frame_optional, /*_sbci_none,*/
	_sbci_value,
	_sbci_frame_index,
	_sbci_offset,
	_sbci_float24,
	_sbci_register,
	_sbci_displaced,
	_sbci_displaced_index,
	_sbci_field,
	_sbci_field_index,
	_sbci_cache_table
};

extern sbcp_opcode_info_data	_sbcp_opcodes[64];
extern sbcp_register_name		_sbcp_register_names[16];

/* ---------- prototypes */

/* at least 32 chars should be allocated for each (valid) string */
/* currently only returns zero or -1 (bad operand) */
EXTERN_API_C( int )
  sbcp_disassemble(_uint8 **position,
	char *opname,		/* can be NULL */
	char *operands,		/* can be NULL */
	char *comment,		/* can be NULL */
	long displacement);	/* for offsets in code */

# ifdef PRAGMA_IMPORT_OFF
#  pragma import off
# elif PRAGMA_IMPORT
#  pragma import reset
# endif

# ifdef __cplusplus
}
# endif
#endif /*!defined(__SBCP_CORE_H)*/
