/*
 sbcp_debug.c
  stack-based core prototype debugging section (optional)
  
  For sbcp version 0.1.7

  Created: January 26, 2007 3:09 AM EDST
  Revised: January 12, 2008 9:30 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Friday, January 26, 2007 3:11:49 AM


[
  Current style of disassembly:
	do
	{
		sbcp_disassemble(&position,buf,buf+32,buf+64,position-code);
		fprintf(stdout,"%s\t%-19s|%s\n", buf,buf+32,buf+64);
	} while(--remaining!=0);
]
*/

#include <stdio.h>
#include <string.h>
#include <fp.h>

#include "sbcp_core.h"
#include "sbcp_utils.h"
/*#include "sbcp_lib.h"*/

#ifdef env68k
# pragma segment sbcp
#endif

/* ---------- globals */

/* refer to sbcp_core.h for operand types (what appears after each string here) */
sbcp_opcode_info_data	_sbcp_opcodes[64] =
{
	"nop" "\0\1",
	"ads" "\0\1",
	"pshi" "\0\4",
	"pshn" "\0\3",
	"nea" "\0\3",
	"exg" "\0\2",
	"stop" "\0\1",
	"ret" "\0\1",

	"pshr" "\0\2",
	"popr" "\0\2",
	"pshd" "\0\2",
	"popd" "\0\2",
	"orc" "\0\2",
	"andc" "\0\2",
	"addc" "\0\2",
	"cmpd" "\0\2",

	"rdd" "\0\5",
	"rds" "\0\5",
	"rdb" "\0\5",
	"rdbz" "\0\5",
	"wdd" "\0\5",
	"wds" "\0\5",
	"wdb" "\0\5",
	"wdbz" "\0\5",

	"rddu" "\0\5",
	"rdsu" "\0\5",
	"rdbu" "\0\5",
	"rdfe" "\0\11",
	"wddu" "\0\5",
	"wdsu" "\0\5",
	"wdbu" "\0\5",
	"wdfe" "\0\11",

	"or" "\0\2",
	"and" "\0\2",
	"exor" "\0\2",
	"snz",
	"sge",
	"sle",
	"slt",
	"sgt",

	"add" "\0\2",
	"mul" "\0\2",
	"div" "\0\2",
	"scal" "\0\2",
	"neg" "\0\2",
	"abs" "\0\2",
	"rndz" "\0\2",
	"sqrt" "\0\2",

	"jn" "\0\3",
	"jnz" "\0\3",
	"cn" "\0\3",
	"cnz" "\0\3",
	"iad" "\0\5",
	"iar" "\0\5",
	"iaa" "\0\5",
	"ci" "\0\5",

	"rct" "\0\6",
	"wct" "\0\6",
	"setrg" "\0\7",
	"getrg" "\0\7",
	"setds" "\0\1",
	"setpc" "\0\1",
	"trap" "\0\1",
	"jmp" "\0\1",
};

sbcp_register_name		_sbcp_register_names[16] =
{
	"DS", /* data stack register */
	"LS", /* link stack register */
	"SR", /* status register */
	"PC", /* program counter */

	"DF", /* data floor */
	"DC", /* data ceiling */
	"LF", /* link floor */
	"LC", /* link ceiling */

	"CR", /* cache register - always zero in emulation */
	"CS", /* cache status register - remains constant in emulation */
	"MR", /* map register */
	"VS", /* vector status register - remains constant in emulation */

	"V0", /* vector table 0 (execution vector) - always zero in emulation */
	"V1", /* vector table 1 (exception vector) */
	"V2", /* vector table 2 (currently undefined) */
	"V3", /* vector table 3 (currently undefined) */
};

/* ---------- code */

static _uint32 ufract32_exp2(_uint32 x)
{
	_int64			n;
	_uint32			s, b = 0xDA5C84F0, i = -(x>>30)+3;
	_int32			c;

	x <<= 2, s = mulhi32_u32(x,x);
	b -= mulhi32_u32(s,0xFF8890C8)>>11;
	b = (mulhi32_u32(b,s)>>10)+0x97D3BAEB; /* scale 4 */
	b = ((mulhi32_u32(b,s)>>6)+0xAB8671A2)>>2; /* scale 11 */
	c = (mulhi32_u32((s>>8)+0xD5487003,s)>>5)+0x7BBAB097-b;
	int64_sethl(n,b,0);
	if ((c>>=1)==0)
		return -1;
	int64_divs32(n,c);
	int64_addhl(n,1,0);
	int64_shr(n,i);

	return int64_to_l(n);
}

static int float24_to_str(char *str, _uint32 x)
{
	_uint8			buf[32], *s = 32+buf;
	char			*d = str;
	_uint64			n;
	_uint32			m = x;
	int				exp;

	if (x&0x800000UL)
		*d++ = '-';
	m &= 0x7FFFFFUL, exp = (m>>16)-63;
	m &= 0xFFFFUL, m += 0x10000L;
	mul64_u32(n,exp,0x9A209A85); /* log10(2) "$3FFD 9A209A84 FBCFF798" */
	uint64_half(n);
	/* int64hi(n) is now the decimal exponent, int64hi(n) an exponent of 10 */
	exp = int64hi(n);
	/* convert back into power of two, get exponent of two and multiply */
	m = mulhi32_u32(ufract32_exp2(mulhi32_u32(int64lo(n),0xD49A784C)),m);/*+((signed)int64lo(n)<0),x); -- round and multiply */

	if ((unsigned)(exp+3)<=9)
	{ /* fixed */
		goto end;
	}
end:
	return d-str;
}

#define write_hex02(d,x)				do{_uint32 __x=x; *d++='0',*d++='x', \
	*d++=hexdigits[(__x>>4)&15],*d++=hexdigits[__x&15];}while(0)
#define write_hex04(d,x)				do{_uint32 __x=x; *d++='0',*d++='x', \
	d[3]=hexdigits[__x&15], \
	d[2]=hexdigits[(__x>>=4)&15], \
	d[1]=hexdigits[(__x>>=4)&15], \
	*d=hexdigits[(__x>>=4)&15], d+=4;}while(0)
#define write_hex06(d,x)				do{_uint32 __x=x; *d++='0',*d++='x', \
	d[5]=hexdigits[__x&15], \
	d[4]=hexdigits[(__x>>=4)&15], \
	d[3]=hexdigits[(__x>>=4)&15], \
	d[2]=hexdigits[(__x>>=4)&15], \
	d[1]=hexdigits[(__x>>=4)&15], \
	*d=hexdigits[(__x>>=4)&15], d+=6;}while(0)
#define write_hex08(d,x)				do{_uint32 __x=x; *d++='0',*d++='x', \
	d[7]=hexdigits[__x&15], \
	d[6]=hexdigits[(__x>>=4)&15], \
	d[5]=hexdigits[(__x>>=4)&15], \
	d[4]=hexdigits[(__x>>=4)&15], \
	d[3]=hexdigits[(__x>>=4)&15], \
	d[2]=hexdigits[(__x>>=4)&15], \
	d[1]=hexdigits[(__x>>=4)&15], \
	*d=hexdigits[(__x>>=4)&15], d+=8;}while(0)

int sbcp_disassemble(_uint8 **position,
	char *opname,
	char *operands,
	char *comment,
	long displacement)
{
	static const char	hexdigits[16] = "0123456789ABCDEF";

	const char		*opcode_info;

	_uint8			*opc, g;
	char			*d;
	_uint32			opcode, t, z;

	if (position==NULL || (opc=*position)==NULL)
		return -1;

	/* piece together the remaining bits of the opcode */
#ifdef SBCP_UNALIGNED_RW_BE
	g = *opc, opcode = *((_uint32*)opc)++&0xFFFFFFUL;
#else
	g = opcode = *++opc;
	opcode <<= 8; opcode |= opc[1];
	opcode <<= 8; opcode |= opc[2];
#endif
	if (operands!=NULL)
		operands[0] = 0;
	if (comment!=NULL)
		comment[0] = 0;
	if ((g&0x40) || *(opcode_info=(const char*)((g&0x3F)+_sbcp_opcodes))==0)
		goto bad; /* this opcode is undefined */

	z = ~(long)opcode_info;
	/*z = (long)memchr(opcode_info,0,8);*/
	if ((d=opname)==NULL)
	{
		t = 8;
		while(*opcode_info++!=0 && --t!=0) ; /* get length of the opcode info string */
	} else
	{
		/*opname += sprintf(opname,"%-8s", opcode_info);*/
		t = 7;
		do
		{
			if ((*d++=*opcode_info++)==0)
				{do{*d++ = ' ';} while(--t!=0); break;}
		} while(--t!=0);

		*d++ = ' '+(('X'-' ')&-((g&0x80)!=0));
		*d = 0;
	}

	z += (long)opcode_info;

	if (z>6) /* nothing remaining for operands */
		goto end;
	switch(*opcode_info)
	{
	case 0:
	case 2:
		if ((d=operands)==NULL)
			break;
		if (opcode!=0x800000L)
		{
			if (opcode!=0)
				if ((g&0x80)==0)
					*d++ = '+';
				else
					*d++ = '-', opcode = (-opcode)&0xFFFFFFL;
			write_hex06(d,opcode);
		} else if (*opcode_info==0)
			break;
		*d++ = '(', *d++ = 'D', *d++ = 'S', *d++ = ')', *d = 0;
		break;
	case 1:
		if ((d=operands)==NULL)
			break;
		/*sprintf(operands,"0x%06X", opcode);*/
		write_hex06(d,opcode); *d = 0;
		break;
	case 3:
		if ((d=comment)!=NULL)
		{
			*d++ = ' ';
			write_hex08(d,((signed)(opcode<<8)>>8)+displacement); *d = 0;
		}
		if ((d=operands)==NULL)
			break;
		*d++ = '*';
		if ((g&0x80)==0)
			*d++ = '+';
		else
			*d++ = '-', opcode = (-opcode)&0xFFFFFFL;
		write_hex06(d,opcode); *d = 0;
		break;
	case 4:
		if ((d=operands)!=NULL)
			{*d++ = '#', *d++ = '"'; write_hex06(d,opcode); *d++ = '"', *d = 0;}
		if ((d=comment)==NULL)
			break;
		*d++ = ' '+(('-'-' ')&-((g&0x80)!=0));
		/*t = sprintf(comment,"%.7G", float24_to_double_t(opcode&0x7FFFFFL));*/
		d[float24_to_str(d,opcode&0x7FFFFFL)] = 0;
		break;
	case 5:
		if (g&0x80)
			goto bad;
		if ((d=operands)==NULL)
			break;
		*d++ = '#'; write_hex02(d,g);
		if ((opcode&0x8000L)==0)
			opcode &= 0xFFFFL, *d++ = '+';
		else
			opcode = (-opcode)&0xFFFFL, *d++ = '-';
		write_hex04(d,opcode);
	case 6:
		if ((d=operands)!=NULL)
			break;
		*d++ = '#'; write_hex02(d,g); *d++ = '+'; write_hex04(d,opcode&0xFFFFL);
		break;
	case 7:
		if ((opcode&0xF0FFFFL)!=0)
			goto bad; /* currently an invalid opcode */
		if ((d=operands)!=NULL)
		{
			const char		*s = _sbcp_register_names[g];

			while((*d++=*s++)!=0) ;
			*d++ = ',';
			if ((opcode&0x8000L)==0)
				opcode &= 0xFFFFL, *d++ = '+';
			else
				opcode = (-opcode)&0xFFFFL, *d++ = '-';
			write_hex04(d,opcode);
			*d++ = '(', *d++ = 'D', *d++ = 'S', *d++ = ')', *d = 0;
		}
		break;
	case 8:
		if ((opcode&224)!=0)
			goto bad;
		if (operands==NULL)
			break;
		*d++ = '#'; write_hex02(d,g);
		if ((opcode&0x8000L)==0)
			opcode &= 0xFFFFL, *d++ = '+';
		else
			opcode = (-opcode)&0xFFFFL, *d++ = '-';
		write_hex04(d,opcode);
		*d++ = '{', *d++ = 'U'+(('S'-'U')&-((opcode&0x800000L)!=0)), *d++ = '}';
		break;
	case 9:
		if ((opcode&0xE0L)!=0)
		{
		bad:
			if ((d=opname)!=NULL)
				*d++ = 'd', *d++ = 'c', *d++ = *d++ = *d++ = ' ', *d = 0;
			if (operands!=NULL)
				*operands = 0;
			if ((d=comment)==NULL)
				break;
			*d++ = ' '; write_hex06(d,*--opc);
			break;
		}
		if (operands==NULL)
			break;
		*d++ = '#'; write_hex02(d,g);
		if (((t=opcode)&0x80L)==0)
			t &= 0xFFL, *d++ = '+';
		else
			t = (-opcode)&0xFFL, *d++ = '-';
		write_hex02(d,t);
		*d++ = '{', *d++ = hexdigits[(opcode>>8)&3], *d++ = ':';
		*d++ = 'U'+(('S'-'U')&-((opcode&0x800000L)!=0));
		t = (opcode&=31);
		if ((opcode=opcode*13>>7)!=0) /* faster /=10 */
			*d++ = hexdigits[opcode];
		*d++ = hexdigits[t-=opcode*10];
		*d++ = '}';
		break;
	}
end:
	*position = opc;
	return 0;
}
