/*
 sbcp_dump.c
  ...just try to keep it versatile

 the user should be forewarned that there is NO standard format here,
   being this as for the public domain !

  Created: December 1, 2007 6:43 PM EDST
  Revised: December 4, 2007 6:51 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <ConditionalMacros.h>

#include <stdlib.h>
#include <errno.h>

#include "sbcp_utils.h"

/* current size */
#define SIZE_OF_DUMP_HEADER				20

/* deliberately put a null at the beginning to stir off text parsers */
#define DUMPFILE_SIGNATURE				FOUR_CHAR_CODE('\0sbc')

/* ---------- code */

static void reverse_bytes(_uint8 *d, int z)
{
	_uint8			*s = z+d, c;
	int				x = z;

	if ((x>>=1)<=0) /* ends before the middle byte */
		return;
	do
	{
		c = *--s; *s = *d; *d++ = c;
	} while(--x);
}

long __freadnumber(void *n, int z, FILE *s)
{
	long			c;

	if (n==NULL)
		return 0;

	c = fread(n,1,z,s);
	if (!TARGET_RT_BIG_ENDIAN)
		reverse_bytes(n,z);

	return c;
}

/*
#define _fgetxint16(d,f)        do{int __x; _uint16 __d=0; \
	if((__x=)==EOF)__d=(_uint8)fgetc(f)<<8; __d|=(_uint8)fgetc(f);}while(0)
#define _fgetxint24(d,f)        do{_uint32 __d=0; __d=(_uint8)fgetc(f)<<16; __d|=(_uint8)fgetc(f)<<8; __d|=(_uint8)fgetc(f);}while(0)
#define _fgetxint32(d,f)        do{_uint32 __d=0; __d=(_uint8)fgetc(f)<<24; __d|=(_uint8)fgetc(f)<<16; __d|=(_uint8)fgetc(f)<<8; __d|=(_uint8)fgetc(f);}while(0)
*/

int sbcp_dump_code(_uint8 *code,
	long length,
	char *filepath,
	int pref_zdata,
	int pref_zlink,
	boolean has_attachment)
{
	FILE			*stream;
	_int32			i32;
	int				t = 1;

	if (filepath==NULL)
		return -1;
	if ((stream=fopen(filepath,"w"))==NULL)
		return errno;

	/* for endian compatibility, the output will always be big-endian */

	/* signature (four bytes with null) */
	i32 = DUMPFILE_SIGNATURE;
	fwrite(&i32,4,1,stream);

	/* version */
	_fputxint16(SBCP_VERSION,stream);

	/* setup info (two bytes) */
	fputc(1-*(_uint8*)&t,stream); /* system endian value (doesn't matter much) */
	fputc(has_attachment,stream);

	/* prefered data stack size */
	_fputxint32(pref_zdata,stream);

	/* prefered link stack size (four bytes) */
	_fputxint32(pref_zlink,stream);
/*0x10*/
	_fputxint32(length,stream);

	if (fwrite(code,length,1,stream)<=0)
		return EOF;

	return 0;
}

int sbcp_load_code(FILE *stream,
	long *out_length,
	_uint8 **out_data,
	sbcp_dump_info *info,
	boolean verbose)
{
	sbcp_dump_info	info_sub;
	_int32			i32, length;
	int				i;

	if (stream==NULL)
		return -1;
	if (info==NULL)
		info = &info_sub;

	/* verify header */
	if (fread(&i32,4,1,stream)<=0)
		return EOF;
	if (i32!=DUMPFILE_SIGNATURE)
		return _sbcd_bad_format;
	if (verbose)
		printf("---- dumpfile header:\n dumpfile signature:       '%-.4s'\n", &i32);
	if (freadnumber(&info->version,stream)<sizeof(info->version))
		return EOF;
	if ((info->version&~0xF)>(SBCP_VERSION&~0xF)) /* ignore the patch digit */
		return _sbcd_new_version;
	if (verbose)
		printf(" recorded sbcp version:     %d\n", info->version);

	info->big_endian = fgetc(stream); /* can skip error check here */
	if ((i=fgetc(stream))<0) /* EOF should be negative */
		return i;
	info->has_attachment = i;

	if (verbose)
		printf(" recorded system endian:    %u (%s)\n", info->big_endian,info->big_endian==0 ? "little":"big");

	if (freadnumber(&info->zdata,stream)<sizeof(info->zdata)
			|| freadnumber(&info->zlink,stream)<sizeof(info->zlink)
			|| freadnumber(&length,stream)<sizeof(length))
		return EOF;

	if (verbose)
		printf(" prefered data stack size:  %d\n prefered link stack size:  %d\n", info->zlink);

	if ((info->offset_to_code=ftell(stream))<0)
		;/*info->offset_to_code = SIZE_OF_DUMP_HEADER; -- don't hide the fact */

	if (out_length!=NULL)
		*out_length = length;
	if (out_data!=NULL && (*out_data=malloc(length))==NULL)
		return _sbcd_memory_error;

	return 0;
}
