/*
 sbcp_execution.c
  stack-based core prototype execution section

 Version: 0.1.7
 Created: January 23, 2007 1:30 AM EDST
 Revised: January 13, 2008 10:34 PM EDST

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.

Tuesday, November 27, 2007 5:06 PM
A.P.: removing NULL and ==mr checking of emulated PC (unnecessary - only reduced performance).
6:27 PM
  replaced `byte' with `UInt8'.
Thursday, November 29, 2007 1:53 PM
  replacing MacOS `xInt' with [custom] `_xint'.
Thursday, November 30, 2007 4:12 PM
  added _insx_trapv.  now removing automatic the break (seemed to only reduce performance).
Monday, December 3, 2007 8:42 AM
  _err_bad_stack_position divided into _err_stack_overflow/_err_stack_underflow.
Friday, December 21, 2007 10:49 AM
  add, mul, div, and scal are getting a frame index for christmas.
Saturday, January 12, 2008 6:00 PM
  to simplify execution (a bit much code for just 63 instructions), opcodes are now uniform.
*/

#include <ConditionalMacros.h>

#include <limits.h>
#include <fp.h>

#include "sbcp_core.h"
#include "sbcp_utils.h"
/*#include "sbcp_lib.h"*/

#ifdef env68k
# pragma segment sbcp
#endif

#define INCREMENT_LINE_COUNTER			1

/* ---------- read/write */

#if _ERM
 /* copy byte-reversed */
# define write_double(d,s)				do{double __x=s; _int8*__s=7+(_int8*)&__x,\
	*__d=(_int8*)(d); *__d++=*__s--,*__d++=*__s--,*__d++=*__s--,*__d++=*__s--; \
	*__d++=*__s--,*__d++=*__s--,*__d++=*__s--,*__d++=*__s--;}while(0)
# define write_single(d,s)				do{float __x=s; _int8*__s=3+(_int8*)&__x,\
	*__d=(_int8*)(d); *__d++=*__s--,*__d++=*__s--,*__d++=*__s--,*__d++=*__s--;}while(0)
#else
# define write_double(d,s)				do{*(double*)(d)=0.0+s;}while(0)
# define write_single(d,s)				do{*(float*)(d)=0.0+s;}while(0)
#endif

static _uint32 read_ufield(_uint8 *s, int offset, int width)
{
	if ((width&=31)==0)
		width = 32;
	s += offset>>3; offset &= 3;

	return ((((s[0]<<24)|(s[1]<<16)|(s[2]<<8)|s[3])<<offset)|(s[4]>>(7-offset)))>>(32-width);
}
static _int32 read_sfield(_uint8 *s, int offset, int width)
{
	if ((width&=31)==0)
		width = 32;
	s += offset>>3; offset &= 3;

	return (_int32)((((s[0]<<24)|(s[1]<<16)|(s[2]<<8)|s[3])<<offset)|(s[4]>>(7-offset)))>>(32-width);
}
static void write_field(_uint8 *s, int offset, int width, _uint32 value)
{
	_uint32			mask;

	if ((width&=31)==0)
		width = 32;
	s += offset>>3; offset &= 3;

	/* because `offset' may nudge the pointer into the fourth byte, we can't do it
	  all at once in 32 bits; the maximum width across byte boundaries is 39 bits */
	value <<= 32-width; /* rise it to the top */
	*s++ = (*s&~((((-128)>>width)&255)>>offset))|(value>>24+offset);

	/* the rest is done only if `width' is large enough */
	if ((width-=8)<=0)
		{width = 0; return;}

	value <<= 8; value >>= offset, --width;
	mask = (_uint32)((long)0x80000000L>>width)>>offset;
	*s++ = (*s&~(mask>>24))|(value>>24);
	*s++ = (*s&~((mask>>16)&255))|((value>>16)&255);
	*s++ = (*s&~((mask>>8)&255))|((value>>8)&255);
	*s = (*s&~(mask&255))|(value&255); /* maximum width here is seven bits */
}

#if 0/*def env68k*/
 /* already built-in as single instructions */
# pragma parameter __D2 _read_ufield(__A0, __D0, __D1)
  static _uint32 _read_ufield(_uint8 *s, int offset, int width) = {0xE9D0, 0x2821};

# pragma parameter __D2 _read_sfield(__A0, __D0, __D1)
  static _int32 _read_sfield(_uint8 *s, int offset, int width) = {0xEBD0, 0x2821};

# pragma parameter _write_field(__A0, __D0, __D1, __D2)
  static void _write_field(_uint8 *s, int offset, int width, _uint32 value) = {0xEFD0, 0x2821};

# define read_ufield(d,o,w,v)			_read_ufield(d,o,w,v)
# define read_sfield(d,o,w,v)			_read_sfield(d,o,w,v)
# define write_field(d,o,w,v)			_write_field(d,o,w,v)
#endif

#define write_dfield(d,o,w,v,f)			do{_uint32 __tmpv; if(f) { \
	__tmpv=(_int32)(v);} else {__tmpv=v;} write_field(d,o,w,__tmpv);}while(0)

#define CHECK_DP(p,u,o)					do{if(p>state->dc-u||p<state->df+o) \
	{t=(p>state->dc-u) ? _ev_stack_underflow:_ev_stack_overflow; \
	tl=(_uint8*)p-state->base_pointer; goto lexception;}}while(0)

#define CHECK_DS(u,o)					CHECK_DP(state->ds,u,o)

#define CHECK_LP(v,u,o)					do{if(v>state->lc-u) { \
	t=_ev_stack_underflow,tl=(_uint8*)state->ls-state->base_pointer; goto lexception;} \
	else if(state->ls<state->lf+o) {t=_ev_stack_overflow, \
		tl=(_uint8*)state->ls-state->base_pointer; goto lexception;}}while(0)

#define IS_REAL_ADDRESS(s,r)			do{long int __tmpl=(_uint8*)(s)-state->base_pointer; \
	||(unsigned)(__tmpl>>=16)>state->page_count||(state->mr[__tmpl]&127)==0) {r;}}while(0)
#define GET_REAL_ADDRESS(d,s,r)			do{_int64 __tmpq=s; d=NULL; \
	int64_shr(__tmpq,16); if(uint64_u32ge(__tmpq,state->page_count) /* Int64N(__tmpq)||*/ \
		||(state->mr[int64_to_l(__tmpq)]&127)==0) {r; break;} \
	d=state->base_pointer+int64_to_l(s);}while(0)

/* these also do a second routine for the case that the mapped page isn't writable */
#define GET_REAL_WADDRESS(d,s,r,w)		do{_int64 __tmpq=s; _uint8 __tmpb; d=NULL; \
	int64_shr(__tmpq,16); if(uint64_u32ge(__tmpq,state->page_count) \
		||((__tmpb=state->mr[int64_to_l(__tmpq)])&127)==0) {r; break;} \
	if((__tmpb&128)==0) {w;} d=state->base_pointer+int64_to_l(s);}while(0)
#define IS_REAL_WADDRESS(s,r,w)			do{long int __tmpl; _uint8 __tmpb; \
	__tmpl=(_uint8*)(s)-state->base_pointer; \
	if((unsigned)(__tmpl>>=16)>state->page_count||((__tmpb=state->mr[__tmpl])&127)==0) \
	{r; break;} if((__tmpb&128)==0) {w;}}while(0)

#define GET_FRAMEINDEX(i,d)				do{if((i)==(long)0xFF800000) \
	{CHECK_DS(1,0); d=*dp; ++state->ds;} else {double *__tmpx; \
	CHECK_LP(state->ls,0,1); __tmpx=(i)+(double*)(state->base_pointer+ \
	int64_to_l(state->ls->return_ds)); CHECK_DP(__tmpx,0,0); d=*__tmpx;}}while(0)

#define SET_FRAMEINDEX(i,s)				do{if((i)==(long)0xFF800000) {CHECK_DS(0,1); \
	dp=--state->ds; *dp=s;} else {double *__tmpx; CHECK_LP(state->ls,0,1); \
	__tmpx=(i)+(double*)(state->base_pointer+int64_to_l(state->ls->return_ds)); \
	CHECK_DP(__tmpx,0,0); *__tmpx=s;}}while(0)

#define FRAMEINDEX_HIOPS(i,d,x)			do{if((i)==(long)0xFF800000) \
	{CHECK_DS(1,0); *(_uint32*)dp x,d=*dp;} else {double *__tmpx; \
	CHECK_LP(state->ls,0,1); __tmpx=(i)+(double*)(state->base_pointer+ \
	int64_to_l(state->ls->return_ds)); CHECK_DP(__tmpx,0,0); *(_uint32*)__tmpx x, \
	d=*(_uint32*)__tmpx;}}while(0)

#define setsign_data(d,s)				do{((_int8*)&(d))[_ERM*7]&=0x7F; ((_int8*)&(d))[_ERM*7]|=s;}while(0)
#define couplesign_data(a,b)			((((_int8*)&(a))[_ERM*7]^((_int8*)&(b))[_ERM*7])&0x80)

static _uint8 sgn_data(double x)
{
	return ((_int8*)&x)[_ERM*7]&0x80;
}

static _uint32 exp_data(double x)
{
	return int64hi(*(_int64*)&x)&0x7FF00000L;
}

static _uint32 man_data(double x)
{
	return (int64hi(*(_int64*)&x)&0x000FFFFFL)|int64lo(*(_int64*)&x);
}

static _uint32 nan_data(double x)
{
	return int64hi(*(_int64*)&x)&0x000FFFFFL;
}

#ifdef env68k /* doubles'll always be in big endian data form */
# define sgn_data(d)					(0x80&*(_uint8*)&(d))
# define exp_data(d)					(0x7FF00000L&*(_uint32*)&(d))
# define man_data(d)					((0x000FFFFFL&*(_uint32*)&(d))|((_uint32*)&(d))[1])
# define nan_data(d)					(0x000FFFFFL&*(_uint32*)&(d))
#endif

/* ---------- private code */

static int read_pointer(const sbcp_state *state,
	_int64 pointer,
	int delta,
	_int64 *result)
{
	_uint8			*real_address;

	/* first, test to see if the data in emulated space is actually mapped */
	int64_adds32(pointer,delta);
	GET_REAL_ADDRESS(real_address,pointer,return ~_ev_access_fault);

	if (result!=NULL)
		*result = *(_int64*)real_address;
	return 0;
}

static int do_exception(sbcp_state *state, int type, double data, int err)
{
	_int64			new_position;
	long int		tl;
	_uint32			old_sr = state->sr;
	int				what = type, appending_data_count = 0;

	switch(what)
	{
	case _ev_stack_overflow:
		if ((state->sr&(long)0x80000000)==0)
		{
			/* over/underflow exceptions are not generated when bit 31 is zero
			  because the floor and ceiling values are ignored */
			return _err_stack_overflow;
		}
		/* over/underflow exceptions don't affect DS here */
		break;
	case _ev_stack_underflow:
		if ((state->sr&(long)0x80000000)==0)
			return _err_stack_underflow;
		break;
	case _ev_access_fault:
	case _ev_write_fault:
	case _ev_illegal_opcode:
	case _ev_illegal_address:
	case _ev_outside_frame:
	case _ev_non_integeral:
		appending_data_count = 1;
		break;
	}

	state->sr |= 0x40000000UL;
	if (state->ls>state->lf)
	{
		sbcp_link	*le = --state->ls;

		tl = ((_uint8*)state->pc)-state->base_pointer; int64_setl(le->return_pc,tl);
		tl = ((_uint8*)state->ds)-state->base_pointer; int64_setl(le->return_ds,tl);
	} else /* sucks if we can't link */
		what = _ev_execution_fault;

	if (what==_ev_execution_fault || state->ds<=state->df+++appending_data_count)
	{
		/*state->sr &= ~(long)0x80000000;*/
		what = _ev_execution_fault;
	} else
		*--(state->ds) = what;

	if ((err=read_pointer(state,state->v1,what*8,&new_position))!=0)
	{ /* something really, really bad just happened */

		/* execute internal flash-ROM code? */
		return err;
	}

	GET_REAL_ADDRESS(state->pc,new_position,break);
	if ((state->pc==NULL || state->pc==state->base_pointer) && err==0)
		if ((state->sr&0x20000000)==0) /* shouldn't have more than one pending exception */
		{
			state->sr |= 0x20000000; /* not doing this could be fatal (real recursion) */
			err = do_exception(state,what,data,type);
		} else
			err = ~what; /* force the error */

	/* write the information regarding the exception */
	switch(what)
	{
	case _ev_access_fault:
	case _ev_write_fault:
	case _ev_illegal_opcode:
	case _ev_illegal_address:
	case _ev_outside_frame:
	case _ev_non_integeral:
		*--(state->ds) = data;
	}

	*--(state->ds) = old_sr;
	state->sr |= 0x20000000;

	return err;
}

static int read_instance_data(sbcp_state *state,
	_uint32 opcode,
	long int tl,
	_int64 pointer)
{
	double			v = 0;
	_uint8			*tp;
	int				t, err = 0;

	/**--state->ds = 0;*/
	int64_adds32(pointer,(signed)(opcode<<16)>>16);
	t = 0; GET_REAL_ADDRESS(tp,pointer,t=1);
	if (t!=0)
		{tl = tp-state->base_pointer; t = _ev_access_fault; goto lexception;}
	switch(((opcode>>23)&0xFF)-_ins_rdd*2)
	{
	case (_ins_rdd-_ins_rdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_rdd-_ins_rdd)*2:
		v = *(double*)tp;
		break;
	case (_ins_rds-_ins_rdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_rds-_ins_rdd)*2:
		v = *(float*)tp;
		break;
	case (_ins_rdb-_ins_rdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_rdb-_ins_rdd)*2:
		v = *(_int8*)tp;
		break;
	case (_ins_rdbz-_ins_rdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_rdbz-_ins_rdd)*2:
		v = *tp;
		break;
	case _ins_rdfd-_ins_rdd)*2:
		{
			double		dd;

			dd = state->ds[-1]; t = dd;
		}
		if (t<=0)
			break; /* non-positive width; result is always zero */
		if (t>=32) /* field-decode instructions clip to [0,32] */
			t = 0;
		if (opcode&0x00800000L)
			v = read_sfield(tp,tl,t);
		else
			v = read_ufield(tp,tl,t);
		break;
	lexception:
		v = tl;
	exception:
		return do_exception(state,t,v,0);
	}
	state->rs = v;

	return err;
}

static int write_instance_data(sbcp_state *state,
	_uint32 opcode,
	long int tl,
	_int64 pointer)
{
	double			v = state->rs;
	_uint8			*tp;
	int				t, err = 0;

	int64_adds32(pointer,(signed)(opcode<<16)>>16);
	t = 0; GET_REAL_WADDRESS(tp,pointer,t=1,t=2);
	if (t!=0)
	{
		tl = tp-state->base_pointer;
		t = (t==2)*(_ev_write_fault-_ev_access_fault)+_ev_access_fault;
		goto lexception;
	}
	switch(((opcode>>23)&0xFF)-_ins_wdd*2)
	{
	case (_ins_wdd-_ins_wdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_wdd-_ins_wdd)*2:
		*(double*)tp = v;
		break;
	case (_ins_wds-_ins_wdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_wds-_ins_wdd)*2:
		*(float*)tp = v;
		break;
	case (_ins_wdb-_ins_wdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_wdb-_ins_wdd)*2:
		*(int8*)tp = v;
		break;
	case (_ins_wdbz-_ins_wdd)*2+1:
		state->instance[(opcode>>16)&127] = pointer;
	case (_ins_wdbz-_ins_wdd)*2:
		*tp = v;
		break;
	case (_ins_wdfd-_ins_wdd)*2:
		if ((t=state->zr)<=0)
			break;
		if (t>=32) /* field-decode instructions clip to [0,32] */
			t = 0;
		write_dfield(tp,tl,t,v,opcode&0x00800000L);
		break;
	lexception:
		v = tl;
	exception:
		err = do_exception(state,t,v,0);
	}

	return err;
}

static int getset_register(sbcp_state *state, int w)
{
	double			v;
	_uint8			*tp;
	long int		tl = 0;
	int				err = 0, t;

	if (w&0x00100000L)
	{ /* set */
		/*{CHECK_DS(0,1); v = *state->ds; ++state->ds;}*/
		v = state->rs+(_int16)w;
		if ((state->sr&0x40000000L)==0)
			{t=_ev_privileged; goto exception;}
		if (v!=floor(v))
			{t=_ev_non_integeral; goto exception;}
		tp = (_uint8*)(long int)v;
	}

	/* for protection reasons, the floor, ceiling and map values are unchanged */
	switch(w>>16)
	{
	case 0: /* set DS */
		tp = (long int)tp+state->base_pointer;
		CHECK_DP((double*)tp,0,0);
		state->ds = (double*)tp;
		break;
	case 1: /* set LS */
		tp = (long int)tp+state->base_pointer;
		CHECK_LP((sbcp_link*)tp,0,0);
		state->ls = (sbcp_link*)tp;
		break;
	case 2: state->sr = (_uint32)tp; break; /* set SR */
	case 3: state->pc = (long int)tp+state->base_pointer; break; /* set PC */
	case 4: /* set DF */
	case 5: /* set LF */
	case 6: /* set DC */
	case 7: /* set LC */
	case 13: /* set V1 */
		return _err_sensitive_register;
	case 16: tl = (_uint8*)state->ds-state->base_pointer; break; /* get DS */
	case 17: tl = (_uint8*)state->ls-state->base_pointer; break; /* get LS */
	case 18: tl = state->sr; break; /* get SR */
	case 19: tl = state->pc-state->base_pointer; break; /* get PC */
	case 20: tl = (_uint8*)state->df-state->base_pointer; break; /* get DF */
	case 21: tl = (_uint8*)state->lf-state->base_pointer; break; /* get LF */
	case 22: tl = (_uint8*)state->dc-state->base_pointer; break; /* get DC */
	case 23: tl = (_uint8*)state->lc-state->base_pointer; break; /* get LC */
	case 25: tl = state->fpr; break; /* get FPR */
	case 26: tl = state->mr-state->base_pointer; break; /* get MR */
	case 27: tl = state->vs; break; /* get VS */
	case 29: tl = int64_to_l(state->v1); break; /* get V1 */

	/*default:
		t = _ev_illegal_opcode;*/
	lexception:
		v = tl;
	exception:
		err = do_exception(state,t,v,0);
	}

	if (!(w&0x00100000L))
		state->rs = tl;

	return err;
}

static double fp_data_bitop(double d, double s, int what)
{
#if 0
	/* this implementation does the operation in relative fashion;
	  it was nixed due to how complicated (and expensive) it is */
	_uint64			bd = *(_uint64*)&d, bs = *(_uint64*)&s, tq;
	int				xpd = (int64hi(bd)>>20)&0x07FF;
	int				xps = (int64hi(bs)>>20)&0x07FF;
	boolean			gd = int64n(bd), gs = int64n(bs);

	if (xpd==0x07FF)
		return d; /* destination is NAN/INF */
	if (xps==0x07FF)
		return s;  /* source is NAN/INF */

	/* expose their mantissas for what they really are */
	int64_and32(bd,0x000FFFFFL,-1L);
	int64_and32(bs,0x000FFFFFL,-1L);
	int64_or32(bd,0x00100000L,0L);
	int64_or32(bs,0x00100000L,0L);

	/* raise 'em to the top */
	int64_shl(bd,11);
	int64_shl(bs,11);

	/* make `d' the larger; adjust `s' to match bit placement */
	if (xps>xpd)
	{
		tq = bd; bd = bs; bs = tq; /* swap bd and bs */
		tq = xpd; xpd = xps; xps = tq; /* swap xpd and xps */
		tq = gd; gd = gs; gs = tq; /* swap gd and gs */
	}
	if ((xpd-=xps)>127)
		xpd = 127;
	uint64_shr(bs,xpd);

	if (gd)
		uint64_neg(bd);
	if (gs)
		uint64_neg(bs);

	switch(what)
	{
	case _ins_or: int64_or(bd,bs); break;
	case _ins_and: int64_and(bd,bs); break;
	case _ins_eor: int64_xor(bd,bs); break;
	}

	/*if (Int64N(bd))
		{uint64_neg(bd); bd = !bd;}*/

	xpd += xps;
	*d = scalb(int64_to_x(bd),xpd-63-0x03FF);
#endif
	_int64			bd, bs;

	int64_setx(bd,d);
	int64_setx(bs,s);
	if (((int64hi(bd)>>20)&0x07FF)==0x07FF)
		return d; /* destination is NAN/INF */
	if (((int64hi(bs)>>20)&0x07FF)==0x07FF)
		return s;  /* source is NAN/INF */

	switch(what)
	{
	case _ins_or: int64_or(bd,bs); break;
	case _ins_and: int64_and(bd,bs); break;
	case _ins_exor: int64_xor(bd,bs); break;
	}

	return int64_to_x(bd);
}

/* ---------- public code */

#if !_ERM
# define NAN004							(0+*(double*)__nan004)
#endif

#define check_fpr_status(s)				do{if(((s)->sr>>8)&_fprs_mask&(s)->sr) { \
	if((err=do_exception((s),_ev_math_exception,0,0))!=0) return err; s=(s)->pc;} \
	(s)->sr&=~_fprn_mask;}while(0)

int sbcp_execute_lines(sbcp_state *state,
	long *maximum_lines,
	long int remaining)
{
#if _ERM
	double			NAN004 = __inf()/__inf();
#else
	static UInt32	__nan004[2] = {0x7FF80080, 0}; /* precalculated using nan("004") */
#endif
	_int64			tw;
	double			*dp, ta, tb, tr;
	_uint8			*s, *tp;
	long int		tl;
	long			substitute_count = LONG_MAX, t = 0;
	int				err = 0;

	/* substitute for NULL or contents of zero */
	if (maximum_lines==NULL || *maximum_lines<0)
		maximum_lines = &substitute_count;
	/* check for any invalid pointers (the PC is checked before every read) */
/*	if ((s=state->pc)==(byte*)NULL ...|| state->df==NULL || state->lf==NULL || state->ds==NULL)*/
	if (state==NULL || state->mr==NULL || state->base_pointer==NULL)
		return _err_null_pointer;
	s = state->pc;
	/* check for any other bad pointers */
	if (state->df>state->dc || state->lf>state->lc)
		{err = _err_floor_above_ceiling; return err;}
	tp=(_uint8*)state->df; IS_REAL_WADDRESS(state->df,return ~_ev_access_fault,return ~_ev_write_fault);
	tp=(_uint8*)state->dc; IS_REAL_WADDRESS((_uint8*)state->dc-1,return ~_ev_access_fault,return ~_ev_write_fault);
	tp=(_uint8*)state->lf; IS_REAL_WADDRESS(state->lf,return ~_ev_access_fault,return ~_ev_write_fault);
	tp=(_uint8*)state->lc; IS_REAL_WADDRESS((_uint8*)state->lc-1,return ~_ev_access_fault,return ~_ev_write_fault);
	/* check for bad positions */
	if (state->ds<state->df || state->ls<state->lf)
		{err = _err_stack_overflow; return err;}
	if (state->ds>state->dc || state->ls>state->lc)
		{err = _err_stack_underflow; return err;}
	if (*maximum_lines<=0)
		return err; /* err is still valid */
	tr = 0;
	do
	{
		_uint8		c;

		/* update the state */
		state->pc = s; dp = state->ds;
		/*if (s==state->base_pointer) -- equivalent of NULL in the emulated environment*/
		if ((tl=s-state->base_pointer)&3) /* odd address */
			{tp = s, tl = tp-state->base_pointer, t = _ev_illegal_address; goto lexception;}*/
		if ((unsigned)(tl>>=16)>state->page_count || (state->mr[tl]&127)==0)
			{tp = s, tl = tp-state->base_pointer, t = _ev_access_fault; goto lexception;}

		c = *s;
		t = read_uint32(s);
		t &= 0x7FFFFFFFL;
	reset:
		switch(c)/*&0x7F)*/
		{
		case _ins_nop:
			t = (signed)(t<<8)>>8;
#if INCREMENT_LINE_COUNTER
			if (t>0)
				*maximum_lines -= t;
#endif
			continue;
		case _ins_ads:
			dp += (signed)(t<<8)>>8;
			CHECK_DP(dp,0,0);
			CHECK_LP(state->ls,0,1);
			if ((tl=(_uint8*)dp-state->base_pointer)>=int64_to_l(state->ls->return_ds))
				{t = _ev_outside_frame; goto lexception;}
			state->ds = dp;
			continue;
		case _ins_pshi:
			t &= 0xFFFFFFL;
			CHECK_DS(0,1); --state->ds;
			float24_to_double(t,dp-1);
			continue;
		case _ins_pshn: /* this used to be "push zero" */
			t = (signed)(t<<8)>>8;
			CHECK_DS(0,1); --state->ds;
			write_double(dp-1,*(double*)(t-4+s));
			continue;
		case _ins_nea:
			t = (signed)(t<<8)>>8;
			tl = s-state->base_pointer-4+t; state->rs = tl;
			continue;
		case _ins_exg:
			t = (signed)(t<<8)>>8;
			if (t==(long)0xFF800000)
				{CHECK_DS(1,0); tr = dp[0]; dp[0] = dp[1]; dp[1] = tr; continue;}
			GET_FRAMEINDEX(t,tr);
			SET_FRAMEINDEX(t,dp[0]);
			dp[0] = tr;
			continue;
		case _ins_stop:
			remaining = 1; /* will decrement to zero */
			*maximum_lines = -1; /* to signal that execution has finished */
			continue;
		case _ins_ret:
			CHECK_LP(state->ls,1,0);
			state->pc = state->base_pointer+int64_to_l(state->ls->return_pc);
			/* DS will be checked when used */
			state->ds = (double*)(state->base_pointer+int64_to_l(state->ls->return_ds));
			++state->ls;
			continue;
		case _ins_pshr:
			t = (signed)(t<<8)>>8;
			SET_FRAMEINDEX(t,state->rs);
			continue;
		case _ins_popr:
			t = (signed)(t<<8)>>8;
			GET_FRAMEINDEX(t,state->rs);
			continue;
		case _ins_pshd:
			t = (signed)(t<<8)>>8;
			CHECK_DS(1,1); --state->ds;
			if (t==(long)0xFF800000)
				{dp[-1] = tr = dp[0]; goto test_math;}
			GET_FRAMEINDEX(t,dp[0]);
			continue;
		case _ins_popd:
			t = (signed)(t<<8)>>8;
			CHECK_DS(1,0); ++state->ds;
			if (t==(long)0xFF800000)
				{tr = dp[0]; goto test_math;}
			SET_FRAMEINDEX(t,dp[0]);
			continue;
		case _ins_orc:
		case _ins_andc:
			t = (signed)(t<<8)>>8;
			check_fpr_status(state);
			GET_FRAMEINDEX(t,ta);
			tr = fp_data_bitop(ta,state->rs,c);
			SET_FRAMEINDEX(t,tr);
			/*state->rs = -(tr==0);*/
			goto test_math;
		case _ins_addc:
			t = (signed)(t<<8)>>8;
			check_fpr_status(state);
			GET_FRAMEINDEX(t,ta);
			tr = state->rs+ta;
			SET_FRAMEINDEX(t,tr);
			/*state->rs = -(tr==0);*/
			goto test_math;
		case _ins_cmpd:
			t = (signed)(t<<8)>>8;
			check_fpr_status(state);
			GET_FRAMEINDEX(t,tr);
			state->rs = tr = relation(tr,state->rs);
			goto test_math;
		case _ins_rdfd:
			CHECK_DS(2,0); state->ds += 2;
			if (1)
			{
				_int64		dw;

				tw = state->instance[(t>>16)&127];
				int64_setx(dw,dp[1]); tl = int64lo(dw)&7; int64_shr(dw,3);
				int64_add(tw,dw);
			} else
			{
		case _ins_rdd:
		case _ins_rds:
		case _ins_rdb:
		case _ins_rdbz:
				/*CHECK_DS(0,1);*/
				tw = state->instance[(t>>16)&127];
			}
			if ((err=read_instance_data(state,t,tl,tw))!=0)
				return err;
			continue;
		case _ins_rdfe:
			if (t&0xE0)
				break;
			/*CHECK_DS(0,1); --state->ds; dp[-1] = 0;*/
			tw = state->instance[(t>>16)&127]; int64_adds32(tw,(signed)(t<<16)>>27);
			GET_REAL_ADDRESS(tp,tw,tl=1);
			if (tl!=0)
				{t = _ev_access_fault, tl = tp-state->base_pointer; goto lexception;}
			if (t&0x00800000L)
				t = read_sfield(tp,(t>>8)&3,t&31);
			else
				t = read_ufield(tp,(t>>8)&3,t&31);
			state->rs = t;
			continue;
		case _ins_wdfd:
			if (1)
			{
				_int64		dw;

				tw = state->instance[(t>>16)&127];
				int64_setx(dw,state->fr); tl = int64lo(dw)&7; int64_shr(dw,3);
				int64_add(tw,dw);
			} else
			{
		case _ins_wdd:
		case _ins_wds:
		case _ins_wdb:
		case _ins_wdbz:
				tw = state->instance[(t>>16)&127];
			}
			if ((err=write_instance_data(state,t,tl,tw))!=0)
				return err;
			continue;
		case _ins_wdfe:
			if (t&0xE0)
				break;
			tw = state->instance[(t>>16)&127]; int64_adds32(tw,(signed)(t<<16)>>27);
			GET_REAL_WADDRESS(tp,tw,tl=1,tl=2);
			if (tl!=0)
				if (tl==2)
					{t = _ev_write_fault, tl = tp-state->base_pointer; goto lexception;}
				else
					{t = _ev_access_fault, tl = tp-state->base_pointer; goto lexception;}
			write_dfield(tp,(t>>8)&3,t&31,state->rs,t&0x00800000L);
			continue;
		case _ins_or:
		case _ins_and:
		case _ins_exor:
			t = (signed)(t<<8)>>8;
			GET_FRAMEINDEX(t,ta); CHECK_DS(1,0); check_fpr_status(state);
			tr = fp_data_bitop(ta,dp[0],c);
			goto test_math;
		case _ins_snz:
		case _ins_sge:
		case _ins_sle:
		case _ins_slt:
		case _ins_sgt:
			{
				static _uint8	scn_invert[] = {1, 1, 1, 0, 0};
				static _uint8	scn_match[] = {_fprn_zero, _fprn_minus, _fprn_zero|_fprn_minus, _fprn_minus, 0};

				CHECK_DS(0,0);
				if ((t=(signed)(t<<8)>>8)==0xFF800000)
					tl = state->fpr;
				else
					GET_FRAMEINDEX(t,tl);
				t = c+-_ins_snz;
				t = scn_invert[t]^(_fprn_zero*((tl&(_fprn_zero|_fprn_minus))==scn_match[t]));
				state->fpr &= ~_fprn_zero, state->fpr |= t;
			}
			continue;
		case _ins_add:
			t = (signed)(t<<8)>>8;
			GET_FRAMEINDEX(t,tr); CHECK_DS(1,0); check_fpr_status(state);
			if ((tr+=dp[0])==0)
			{
				if (dp[0]!=0)
					state->fpr |= _fprn_underflow|_fprs_underflow;
			} else if (man_data(tr)==0 && exp_data(tr)==0x7FF00000L
					&& man_data(dp[0])!=0 || exp_data(dp[0])!=0x7FF00000L)
				state->fpr |= _fprn_overflow|_fprs_overflow;
			goto set_math;
		case _ins_mul:
			t = (signed)(t<<8)>>8;
			GET_FRAMEINDEX(t,ta); CHECK_DS(1,0); check_fpr_status(state);
			if ((tr=ta*dp[0])==0)
			{
				if (dp[0]!=0)
					state->fpr |= _fprn_underflow|_fprs_underflow;
			} else if (man_data(tr)==0 && exp_data(tr)==0x7FF00000L
					&& man_data(dp[0])!=0 || exp_data(dp[0])!=0x7FF00000L)
				state->fpr |= _fprn_overflow|_fprs_overflow;
			goto set_math;
		case _ins_div:
			t = (signed)(t<<8)>>8;
			GET_FRAMEINDEX(t,ta); CHECK_DS(1,0); check_fpr_status(state);
			if (ta==0)
			{ /* because some systems won't hold back, this may have to be done manually */
				_int8		sign = couplesign_data(ta,dp[0]);

				/* in math exceptions, here, DBZ comes before NAN */
				state->fpr |= _fprs_dbz|_fprs_divide_by_zero;
				if (dp[0]==0)
					dp[0] = NAN004;
				else
					dp[0] = __inf();
				setsign_data(dp[0],sign);
				tr = dp[0];
			} else
			{
				if ((tr=dp[0]/ta)==0)
				{
					if (dp[0]!=0)
						state->fpr |= _fprn_underflow|_fprs_underflow;
				} else if (man_data(tr)==0 && exp_data(tr)==0x7FF00000L
						&& man_data(dp[0])!=0 || exp_data(dp[0])!=0x7FF00000L)
					state->fpr |= _fprn_overflow|_fprs_overflow;
			set_math:
				dp[0] = tr;
			}
			goto test_math;
		case _ins_scal:
			t = (signed)(t<<8)>>8;
			CHECK_DS(1,0); ++state->ds;
			GET_FRAMEINDEX(t,ta); check_fpr_status(state);
			if ((tr=scalb(ta,dp[0]))==0)
			{
				if (ta!=0)
					state->fpr |= _fprn_underflow|_fprs_underflow;
			} else if (man_data(tr)==0 && exp_data(tr)==0x7FF00000L
					&& man_data(ta)!=0 || exp_data(ta)!=0x7FF00000L)
				state->fpr |= _fprn_overflow|_fprs_overflow;
			SET_FRAMEINDEX(t,tr);
			goto test_math;
		case _ins_neg:
			t = (signed)(t<<8)>>8;
			FRAMEINDEX_HIOPS(t,tr,^=0x80000000); check_fpr_status(state);
			goto test_math;
		case _ins_abs:
			t = (signed)(t<<8)>>8;
			FRAMEINDEX_HIOPS(t,tr,&=0x80000000); check_fpr_status(state);
			goto test_math;
		case _ins_rndz: /* trunc isn't always reliable here; floor and ceil are used */
			CHECK_DS(0,0); check_fpr_status(state);
			if (dp[0]<0)
				tr = ceil(dp[0]);
			else
				tr = floor(dp[0]);
			dp[0] = tr;
			goto test_math;
		case _ins_sqrt:
			CHECK_DS(0,0); check_fpr_status(state);
			if (!(dp[0]>=0))
				state->fpr |= _fprs_operand_error_a;
			tr = dp[0] = sqrt(dp[0]); /* underflow/overflow is impossible for sqrt */
		test_math:
			state->fpr |= (sgn_data(tr)!=0)*_fprn_minus;
			tl = man_data(tr);
			if ((t=exp_data(tr))!=0x7FF00000L)
				state->fpr |= ((t|tl)==0)*_fprn_zero;
			else
			{
				state->fpr |= _fprn_nan-(tl==0)*_fprn_infinite;
				/*if (nan_data(tr)&0x00400000L)
					state->fpr |= _fprs_signaling_nan;*/
			}
			continue;
		case _ins_jnz:
			if ((state->fpr&_fprn_zero)==0)
				continue;
		case _ins_jn:
			t = (signed)(t<<8)>>8;
			s += t-4;
			continue;
		case _ins_cnz:
			if ((state->fpr&_fprn_zero)==0)
				continue;
		case _ins_cn:
			t = (signed)(t<<8)>>8;
			if (state->ls>state->lf)
			{
				sbcp_link	*le = --state->ls;

				tl = ((_uint8*)state->pc)-state->base_pointer; int64_setl(le->return_pc,tl);
				tl = ((_uint8*)state->ds)-state->base_pointer; int64_setl(le->return_ds,tl);
			} else
			{
				t = _ev_stack_overflow,tl = (_uint8*)state->ls-state->base_pointer;
				goto lexception;
			}
			s += t-4;
			continue;
		case _ins_iad:
			if ((t&=0xFFFFFFL)&0x00800000L)
				break;
			CHECK_DS(0,1); --state->ds;
			tw = state->instance[(t>>16)&127]; int64_adds32(tw,(signed)(t<<16)>>16);
			dp[-1] = int64_to_x(tw);
			continue;
		case _ins_iar:
			if ((t&=0xFFFFFFL)&0x00800000L)
				break;
			CHECK_DS(0,1); ++state->ds;
			int64_setx(tw,dp[0]); int64_adds32(tw,(signed)(t<<16)>>16);
			state->instance[(t>>16)&127] = tw;
			continue;
		case _ins_ci:
			if (t&0x00800000L)
				break;
			if (state->ls>state->lf)
			{
				sbcp_link	*le = --state->ls;

				tl = ((_uint8*)state->pc)-state->base_pointer; int64_setl(le->return_pc,tl);
				tl = ((_uint8*)state->ds)-state->base_pointer; int64_setl(le->return_ds,tl);
			} else
			{
				t = _ev_stack_overflow, tl = (_uint8*)state->ls-state->base_pointer;
				goto lexception;
			}
			tr = 0; tw = state->instance[(t>>16)&127];
			int64_adds32(tw,(signed)(t<<16)>>16);
			GET_REAL_ADDRESS(tp,tw,tr=1);
			if (tr!=0)
				{tl = tp-state->base_pointer, t = _ev_access_fault; goto lexception;}
			/*if (tp!=(_uint8*)NULL && tp!=state->base_pointer)*/
			s = tp;
			continue;
		case _ins_rct:
		case _ins_wct:
			t &= 0xFFFFFFL;
			continue;
		case _ins_setrg:
		case _ins_getrg:
			if (((t&=0xFFFFFFL)&0x00F00000L)!=0)
				break;
			if ((err=getset_register(state,t|=(c&1)<<16))!=0)
				return err;
			continue;
		case _ins_setds:
		case _ins_setpc:
			tl = t&=0xFFFFFFL, tp = state->base_pointer+tl;
			if ((state->sr&0x40000000L)==0)
				{t = _ev_privileged; goto lexception;}
			if (c==_ins_setds)
				{CHECK_DP((double*)tp,0,0); state->ds = (double*)tp;}
			else if (c==_ins_setpc)
			{
				if (tl==0)
					{tl = tp-state->base_pointer, t = _ev_illegal_address; goto lexception;}
				state->pc = tp;
			}
			continue;
		case _ins_trap:
			tl = t&0xFFFFFFL; t = _ev_trap_vector;
			goto lexception;

		default:
			if (t&0x80)
				if (state->sr&0x10000000L)
					{t = _ev_break_vector, tl = 0; goto lexception;}
				else
					{c &= 0x7F; goto reset;}
		}
		t = _ev_illegal_opcode, tl = 0;
	lexception:
		if ((err=do_exception(state,t,tl,0))!=0)
			return err;
		s = state->pc;
#if INCREMENT_LINE_COUNTER
	} while((*maximum_lines>0 && remaining<=0) || --remaining!=0);
#else
	} while(remaining<=0 || --remaining!=0);
#endif
	state->pc = s;

	return err;
}
