/*
 sbcp_lib.h
  stack based core prototype library include

  For sbcp version 0.1.2

  Created: 2007.01.23 15:13 GMT
  Revised: 2007.12.09 01:00 GMT

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
 */

/*
    `@' = source also serves as destination

    standard function calls by the go to instance zero, offset index<<3
*/
enum
{ /* built-in library routines -- offsets 0x0000-0x03FF are reserved */
	/* 0x000000			Name/Description			 Parameters					*/
	_lib_random,	/* random						(none)						*/
					/*	uses the cliff random algorithim; result is [0,1.0]		*/

	_lib_setrseed,	/* set random seed				value						*/
					/*	sets the random seed used in the "rand" routine			*/

	_lib_getrseed,	/* get random seed				value						*/
					/*	returns the random seed used in "rand"					*/

	_lib_copysgn,	/* copy sign					value						*/
					/*	copysign(dp[0],dp[1])									*/

	_lib_floor,		/* floor						value						*/
					/*	floor(dp[0])											*/

	_lib_ceil,		/* ceiling						value						*/
					/*	ceil(dp[0])												*/

	_lib_round,		/* round to nearest				value						*/
					/*	round(dp[0])											*/

	_lib_frexp,		/* fractionalize exponent		@value						*/
					/*	dp[0] = frexp(dp[0],&result)							*/
	/* 0x000008 */
	_lib_min,		/* minimus of values			count,value[]				*/
					/*	result is the value closest to -INF; dp[0] is count		*/

	_lib_max,		/* maximus of values			count,value[]				*/
					/*	result is the value closest to +INF; dp[0] is count		*/

	_lib_dim,		/* absolute difference			value[2]					*/
					/*	dim(dp[0],dp[1])										*/

	_lib_next,		/* next after					value[2]					*/
					/*	nextafter(dp[0],dp[1])									*/

	_lib_pow,		/* raise to power				value[2]					*/
					/*	pow(dp[0],dp[1])										*/

	_lib_logx,		/* base-x logarithm				value[2]					*/
					/*	log2(dp[0])/log2(dp[1])									*/

	_lib_log2,		/* base-2 logarithm				value						*/
					/*	log2(dp[0])												*/

	_lib_logb,		/* binary logarithm				value						*/
					/*	logb(dp[0])												*/
	/* 0x000010 */
	_lib_exp,		/* base e exponentiation		value						*/
					/*	exp(dp[0])												*/

	_lib_expm1,		/* base e-1 exponentiation		value						*/
					/*	expm1(dp[0])											*/

	_lib_log,		/* natural logarithm			value						*/
					/*	log(dp[0])												*/

	_lib_log1p,		/* base e+1 logarithm			value						*/
					/*	log1p(dp[0])											*/

	_lib_cos,		/* cosine						value						*/
					/*	cos(dp[0])												*/

	_lib_sin,		/* sine							value						*/
					/*	sin(dp[0])												*/

	_lib_tan,		/* base e exponentiation		value						*/
					/*	tan(dp[0])												*/

	_lib_atan2,		/* arctangent-2					value[2]					*/
					/*	atan2(dp[0],dp[1]) -- standard atan would be atan2(x,1)	*/
	/* 0x000018 */
	_lib_acos,		/* arccosine					value						*/
					/*	acos(dp[0])												*/

	_lib_asin,		/* arcsine													*/
					/*	asin(dp[0])					value						*/

	_lib_cosh,		/* hyperbolic cosine										*/
					/*	cosh(dp[0])												*/

	_lib_sinh,		/* hyperbolic sine				value						*/
					/*	sinh(dp[0])												*/

	_lib_tanh,		/* hyperbolic tangent			value						*/
					/*	tanh(dp[0])												*/

	_lib_acosh,		/* hyperbolic arccosine			value						*/
					/*	acosh(dp[0])											*/

	_lib_asinh,		/* hyperbolic arcsine			value						*/
					/*	asinh(dp[0])											*/

	_lib_atanh,		/* hyperbolic arctangent		value						*/
					/*	atanh(dp[0])											*/
	/* 0x000020 */
	_lib_erf,		/* error function				value						*/
					/*	erf(dp[0])												*/

	_lib_erfc,		/* complementary erf			value						*/
					/*	erfc(dp[0])												*/

	_lib_gamma,		/* gamma function				value						*/
					/*	gamma(dp[0])											*/

	_lib_lgamma,	/* log-e gamma					value						*/
					/*	lgamma(dp[0])											*/

	_lib_rem,		/* remainder					value[2]					*/
					/*	remainder(dp[0],dp[1])									*/

	_lib_hypot,		/* find hypotenuse				value[2]					*/
					/*	hypot(dp[0],dp[1])										*/

	_lib_compound,	/* compound intrest factor		value[2]					*/
					/*	compound(dp[0],dp[1])									*/

	_lib_annuity,	/* present value fact. annuity	value[2]					*/
					/*	annuity(dp[0],dp[1])									*/
	/* 0x000028 */
	_lib_nanr,		/* NAN index read				value						*/
					/*	returns the Not-A-Number index; -1 if number is real	*/

	_lib_nanx,		/* NAN index make				index						*/
					/*	returns a NAN with index; sign id copied to result		*/

	_lib_setrm,		/* set rounding mode			direction					*/
					/*	sets the rounding mode for "nearby" type functions		*/

	_lib_getrm,		/* get rounding mode			(none)						*/
					/*	returns the current rounding mode						*/

	_ins_setps,		/* set program status			status						*/
					/*	sets the current status in case of program stop			*/

	_ins_getps,		/* get program status			(none)						*/
					/*	returns the current program status						*/

	_lib_setev,		/* set exception vector			index, address				*/
					/*	sets indexed exception vector							*/

	_lib_getev,		/* get exception vector			index, address				*/
					/*	returns the indexed exception vector					*/
	/* 0x000030 */
	_lib_movem,		/* move memory					address, address, size		*/
					/*	copies an address to another; result is error code		*/

	_lib_cmpm,		/* compare memory				address, address, size		*/
					/*	compare two addresses; last inequal byte difference		*/

	_lib_chrmf,		/* char in memory forward		address, value, size		*/
					/*	finds char from start of address; result is offset		*/

	_lib_rchrm,		/* char in memory reverse		address, size, value		*/
					/*	finds char from end of address; result is offset		*/

	_lib_moves,		/* move string					straddr, straddr			*/
					/*	copies a string from address to address					*/

	_lib_cmps,		/* compare string				straddr, straddr			*/
					/*	compare two strings; last inequal byte difference		*/

	_lib_chrsf,		/* char in string forward		straddr, value				*/
					/*	same as "forward char in memory", but in C-string form	*/

	_lib_chrsr,		/* char in string reverse		straddr, value				*/
					/*	same as "reverse char in memory", but in C-string form	*/
	/* 0x000038 */
	_lib_bsearch,	/* STDC bsearch					key, base, count, size, rtn	*/
					/*	bsearch(dp[0],dp[1],dp[2],dp[3],dp[4])					*/

	_lib_qsort,		/* STDC qsort					base, count, size, routine	*/
					/*	qsort(dp[0],dp[1],dp[2],dp[3])							*/

	_lib_getenv,	/* STDC get environment var.	straddr						*/
					/*	getenv(dp[0])											*/

	_lib_system,	/* STDC system status			straddr						*/
					/*	system(dp[0])											*/

	_lib_malloc,	/* STDC memory allocate			size						*/
					/*	malloc(dp[0])											*/

	_lib_calloc,	/* STDC custom malloc			count, elemsize				*/
					/*	calloc(dp[0],dp[1])										*/

	_lib_realloc,	/* STDC reallocate malloc		address, newsize			*/
					/*	realloc(dp[0],dp[1])									*/

	_lib_free,		/* STDC free malloc				address						*/
					/*	free(dp[0])												*/
	/* 0x000040 - remove is deliberately unimplemented, or a.k.a. ... "removed" */
	_lib_open,		/* custom file open				addr: read, write, close	*/
					/*	opens a custom file using the three addresses			*/

	_lib_close,		/* file close					fileaddr					*/
					/*	closes a file											*/

	_lib_write,		/* file read					address, count[2], fileaddr	*/
					/*	fwrite(dp[0],dp[1],dp[2],dp[3])							*/

	_lib_read,		/* file read					address, count[2], fileaddr	*/
					/*	fread(dp[0],dp[1],dp[2],dp[3])							*/

	_lib_setpos,	/* file set position			fileaddr, position, whence	*/
					/*	fsetpos(dp[0],dp[1],dp[2])								*/

	_lib_getpos,	/* file read					fileaddr, @position			*/
					/*	fgetpos(dp[0],dp[1])									*/

	_lib_rewind,	/* file rewind					fileaddr					*/
					/*	rewind(dp[0])											*/

	_lib_flush,		/* file flush					fileaddr					*/
					/*	fflush(dp[0])											*/
	/* 0x000048 */
	_lib_fopen,		/* file open					straddr, mode				*/
					/*	opens a file for a disk file - C-string is name			*/

	_lib_freopen,	/* file reopen					straddr, mode, fileaddr		*/
					/*	freopen version of _lib_fopen							*/

	_lib_ferror,	/* file report any errors		fileaddr					*/
					/*	ferror(dp[0])											*/

	_lib_clearerr,	/* clear file error				fileaddr					*/
					/*	clearerr(dp[0])											*/

	_lib_fgetc,		/* file get char				fileaddr					*/
					/*	fgetc(dp[0])											*/

	_lib_fputc,		/* file put char				char, fileaddr				*/
					/*	fputc(dp[0],dp[1])										*/

	_lib_fgets,		/* file get string				fileaddr					*/
					/*	fgets(dp[0])											*/

	_lib_fputs,		/* file put string				straddr, fileaddr			*/
					/*	fputs(dp[0],dp[1])										*/
	/* 0x000050 */
	_lib_fscanf,	/* file scan format				fileaddr, straddr, ...		*/
					/*	fscanf(dp[0],dp[1],...)									*/

	_lib_fprintf,	/* file print format			fileaddr, straddr, ...		*/
					/*	fprintf(dp[0],dp[1],...)								*/

	_lib_sscanf,	/* string scan format			straddr, straddr, ...		*/
					/*	sscanf(dp[0],dp[1],...)									*/

	_lib_sprintf,	/* string print format			straddr, straddr, ...		*/
					/*	sprintf(dp[0],dp[1],...)								*/
	/* 0x000070 */
	_lib_getglobalv	/* get global var. as value		straddr						*/
		= 0x0070,	/*	getglobalv(dp[0])										*/

	_lib_setglobalv	/* set global var. to value		straddr, number (double)	*/
					/*	setglobalv(dp[0],dp[1])									*/

	_lib_getglobalt	/* get global var. as text		straddr						*/
					/*	getglobalt(dp[0])										*/

	_lib_setglobalt	/* set global var. to text		straddr, straddr			*/
					/*	setglobalt(dp[0],dp[1])									*/
};
