/*
 sbcp_utils.h
  covers the sbcp compiler, debugger and... dumper

  For sbcp version 0.1.6

  Created: 2007.01.26 08:37 GMT
  Revised: 2007.12.23 03:12 GMT

 * THIS SOFTWARE IS NOT COPYRIGHTED
 *
 *  This source code is offered for use in the public domain.  You may use,
 * modify or distribute it freely.  If changes have been made to the original
 * copy, please make it clear.
 *
 *  This code is distributed in the hope that it will be useful but WITHOUT
 * ANY WARRANTY.  ALL WARRANTIES, EXPRESS OR IMPLIED ARE HEREBY DISCLAIMED.
 *  This includes but is not limited to warranties of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.
*/

#ifndef __SBCP_UTILS_H
# define __SBCP_UTILS_H

# ifndef __CONDITIONALMACROS__
#  include <ConditionalMacros.h>
# endif

# include <stdio.h>

# include "sbcp.h"

# if PRAGMA_ONCE
#  pragma once
# endif

# ifdef __cplusplus
   extern "C" {
# endif

# if PRAGMA_IMPORT
#  pragma import on
# endif

/* ---------- constants */

/* Endian Reverse Multiple (fast way of putting things at the end on little endian systems) */
# if TARGET_RT_BIG_ENDIAN
#  define _ERM                          0
# else
#  define _ERM                          1
# endif

/* compiler error codes */
typedef enum
{
    _sbcc_no_error,
    _sbcc_syntax_error,         /* the so-called language shouldn't be that hard to learn */
    _sbcc_memory_error,         /* probably ran out of memory */
    _sbcc_undefined,            /* identifier not defined */
    _sbcc_bad_char,             /* character not supported there */
    _sbcc_no_operand,           /* expected operand(s) not found */
    _sbcc_value_error,          /* could not encode a (finite) value */
    _sbcc_duplicate_label,      /* already a label of that name */
    _sbcc_name_too_long,        /* greater than 255 chars */
    _sbcc_internal_error        /* oh, crap */
} sbcc_error;

/* dump/load error codes */
enum
{
    _sbcd_memory_error      = 1,/* probably ran out of memory */
    _sbcd_bad_format,           /* header isn't recognizable */
    _sbcd_new_version           /* version in header isn't compatible */
};

/* ---------- structures */

typedef struct sbcp_dump_info/*(16+)*/
{
    long int        offset_to_code;
    _int32          zdata, zlink;
    _int16          version;
    boolean         big_endian, has_attachment;
} sbcp_dump_info;

/* ---------- macros */

#if _ERM /*|| defined(THINK_C)*/
# define _encodexint16(d,x)     do{_int8 _x[2]; *(_int16*)_x=x,*((_int8*)d)++=_x[1],*((_int8*)d)++=_x[0];}while(0)
# define _encodexint24(d,x)     do{_int8 _x[4]; *(_int32*)_x=x,*((_int8*)d)++=_x[2],*((_int8*)d)++=_x[1],*((_int8*)d)++=_x[0];}while(0)
# define _encodexint32(d,x)     do{_int8 _x[4]; *(_int32*)_x=x,*((_int8*)d)++=_x[3],*((_int8*)d)++=_x[2],*((_int8*)d)++=_x[1],*((_int8*)d)++=_x[0];}while(0)
# define _encodedouble(d,x)     do{_int8 _x[8]; *(double*)_x=x,*((_int8*)d)++=_x[7],*((_int8*)d)++=_x[6],*((_int8*)d)++=_x[5],*((_int8*)d)++=_x[4],*((_int8*)d)++=_x[3],*((_int8*)d)++=_x[2],*((_int8*)d)++=_x[1],*((_int8*)d)++=_x[0];}while(0)

# define _decodeuint16(s,x)     do{_int8 _x[2]; _x[1]=*((_int8*)s))++,_x[0]=*((_int8*)s)++,x=*(_uint16*)_x;}while(0)
# define _decodeuint24(s,x)     do{_int8 _x[4]; _x[3]=0,_x[2]=*((_int8*)s)++,_x[1]=*((_int8*)s)++,_x[0]=*((_int8*)s)++,x=*(_uint32*)_x;}while(0)
# define _decodeuint32(s,x)     do{_int8 _x[4]; _x[3]=*((_int8*)s)++,_x[2]=*((_int8*)s)++,_x[1]=*((_int8*)s)++,_x[0]=*((_int8*)s)++,x=*(_uint32*)_x;}while(0)
# define _decodesint16(s,x)     do{_int8 _x[2]; _x[1]=*((_int8*)s))++,_x[0]=*((_int8*)s)++,x=*(_int16*)_x;}while(0)
# define _decodesint24(s,x)     do{_int8 _x[4]; _x[3]=*((_int8*)s)++,_x[2]=*((_int8*)s)++,_x[1]=*((_int8*)s)++,x=(*(_int32*)_x)>>8;}while(0)
# define _decodesint32(s,x)     do{_int8 _x[4]; _x[3]=*((_int8*)s)++,_x[2]=*((_int8*)s)++,_x[1]=*((_int8*)s)++,_x[0]=*((_int8*)s)++,x=*(_int32*)_x;}while(0)
# define _decodedouble(s,x)     do{_int8 _x[8]; _x[7]=*((_int8*)s)++,_x[6]=*((_int8*)s)++,_x[5]=*((_int8*)s)++,_x[4]=*((_int8*)s)++,_x[3]=*((_int8*)s)++,_x[2]=*((_int8*)s)++,_x[1]=*((_int8*)s)++,_x[0]=*((_int8*)s)++,x=*(double*)_x;}while(0)

# define _fputxint16(x,f)       do{_int8 _x[2]; *(_int16*)&_x=x,fputc(_x[1],f); fputc(_x[0],f);}while(0)
# define _fputxint24(x,f)       do{_int8 _x[4]; *(_int32*)&_x=x,fputc(_x[2],f); fputc(_x[1],f),fputc(_x[0],f);}while(0)
# define _fputxint32(x,f)       do{_int8 _x[4]; *(_int32*)&_x=x,fputc(_x[3],f); fputc(_x[2],f),fputc(_x[1],f),fputc(_x[0],f);}while(0)
#else
# define _encodexint16(d,x)     do{*((_int16*)d)++=(_int16)(x);}while(0)
# define _encodexint24(d,x)     do{_int8 _x[4]; *(_int32*)_x=x; *((_int8*)d)++=_x[1],*((_int8*)d)++=_x[2],*((_int8*)d)++=_x[3];}while(0)
# define _encodexint32(d,x)     do{*((_int32*)d)++=(_int32)(x);}while(0)
# define _encodedouble(d,x)     do{*((double*)d)++=(double)(x);}while(0)

# define _decodeuint16(s,x)     do{x=*((_uint16*)s)++;}while(0)
# define _decodeuint24(s,x)     do{_int8 _x[4]; _x[0]=0,_x[1]=*((_int8*)s)++,_x[2]=*((_int8*)s)++,_x[3]=*((_int8*)s)++,x=*(_int32*)_x;}while(0)
# define _decodeuint32(s,x)     do{x=*((_uint32*)s)++;}while(0)
# define _decodesint16(s,x)     do{x=*((_int16*)s)++;}while(0)
# define _decodesint24(s,x)     do{_int8 _x[4]; _x[0]=*((_int8*)s)++,_x[1]=*((_int8*)s)++,_x[2]=*((_int8*)s)++,x=(*(_int32*)_x)>>8;}while(0)
# define _decodesint32(s,x)     do{x=*((_int32*)s)++;}while(0)
# define _decodedouble(s,x)     do{x=*((double*)s)++;}while(0)

# define _fputxint16(x,f)       do{_int8 _x[2]; *(_int16*)_x=x,fwrite(_x,2,1,f);}while(0)
# define _fputxint24(x,f)       do{_int8 _x[4]; *(_int32*)_x=x,fwrite(1+_x,3,1,f);}while(0)
# define _fputxint32(x,f)       do{_int8 _x[4]; *(_int32*)_x=x,fwrite(_x,4,1,f);}while(0)
#endif

/* ---------- prototypes */

/* call this to see if a float24 doesn't overflow (underflow still passes) */
EXTERN_API_C( boolean )
  sbcp_float24_range_check(double v);

/* makes a float24 (24 bits, unsigned) from a double; used in sbcp code */
EXTERN_API_C( _uint32 )
  sbcp_make_float24(double v);

EXTERN_API_C( void )
  float24_to_double(_uint32 x, double *v);

/* takes any size and automatically converts to native endian */
EXTERN_API_C( long )
  __freadnumber(void *n,
    int size,
    FILE *stream);

#define freadnumber(n,s)        __freadnumber(n,sizeof(n),s)

/* the user should be forewarned that there is NO standard format here,
  being this is for the public domain ! */
EXTERN_API_C( int )
  sbcp_dump_code(_uint8 *code,
    long length,
    char *filepath,
    int pref_zdata,
    int pref_zlink,
    boolean has_attachment);

/* reads a dump file, and (optionally) loads the code into a new pointer */
EXTERN_API_C( int )
  sbcp_load_code(FILE *file,
    long *length, /* can be NULL */
    _uint8 **out_data, /* can be NULL */
    sbcp_dump_info *out_info,
    boolean verbose); /* pass true to printf the header */

/* parses the number (accepts hexadecimal and octal as well); returns the next
  character after the number, or NULL if the syntax isn't correct */
EXTERN_API_C(const char*)
  sbcp_parse_number(const char *text,
    const char *text_end, /* the first pointer after `text' that isn't valid */
    double *out_value,
    boolean real); /* otherwise the decimal point isn't allowed */

/* compiles the text, and (optionally) makes code in a new pointer*/
EXTERN_API_C( sbcc_error )
  sbcp_compile(const char *text,
    long *length, /* length of text in, length for outdata out */
    _uint8 **out_data, /* can be NULL (syntax check) */
    long *err_offset); /* offset in text where error occured; can be NULL */

# ifdef PRAGMA_IMPORT_OFF
#  pragma import off
# elif PRAGMA_IMPORT
#  pragma import reset
# endif

# ifdef __cplusplus
   }
# endif
#endif /*!defined(__SBCP_UTILS_H)*/
